import pytest

import pyteal as pt
from pyteal import abi
from pyteal.ast.abi.util import substring_for_decoding
from pyteal.ast.abi.tuple import _encode_tuple
from pyteal.ast.abi.bool import _bool_sequence_length
from pyteal.ast.abi.type_test import ContainerType
from pyteal.ast.abi.array_base_test import STATIC_TYPES, DYNAMIC_TYPES

options = pt.CompileOptions(version=5)


def test_StaticArrayTypeSpec_init():
    for elementType in STATIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            assert staticArrayType.value_type_spec() is elementType
            assert not staticArrayType.is_length_dynamic()
            assert staticArrayType._stride() == elementType.byte_length_static()
            assert staticArrayType.length_static() == length

        with pytest.raises(TypeError):
            abi.StaticArrayTypeSpec(elementType, -1)

    for length in range(256):
        staticBytesType = abi.StaticBytesTypeSpec(length)
        assert isinstance(staticBytesType.value_type_spec(), abi.ByteTypeSpec)
        assert not staticBytesType.is_length_dynamic()
        assert staticBytesType._stride() == 1
        assert staticBytesType.length_static() == length

    for elementType in DYNAMIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            assert staticArrayType.value_type_spec() is elementType
            assert not staticArrayType.is_length_dynamic()
            assert staticArrayType._stride() == 2
            assert staticArrayType.length_static() == length

        with pytest.raises(TypeError):
            abi.StaticArrayTypeSpec(elementType, -1)


def test_StaticArrayTypeSpec_str():
    for elementType in STATIC_TYPES + DYNAMIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            assert str(staticArrayType) == "{}[{}]".format(elementType, length)

    for length in range(256):
        assert str(abi.StaticBytesTypeSpec(length)) == f"byte[{length}]"


def test_StaticArrayTypeSpec_new_instance():
    for elementType in STATIC_TYPES + DYNAMIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            instance = staticArrayType.new_instance()
            assert isinstance(
                instance,
                abi.StaticArray,
            )
            assert instance.type_spec() == staticArrayType

    for length in range(256):
        staticBytesType = abi.StaticBytesTypeSpec(length)
        instance = staticBytesType.new_instance()
        assert isinstance(instance, abi.StaticBytes)
        assert instance.type_spec() == staticBytesType


def test_StaticArrayTypeSpec_eq():
    for elementType in STATIC_TYPES + DYNAMIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            assert staticArrayType == staticArrayType
            assert staticArrayType != abi.StaticArrayTypeSpec(elementType, length + 1)
            assert staticArrayType != abi.StaticArrayTypeSpec(
                abi.TupleTypeSpec(elementType), length
            )

    for length in range(256):
        staticBytesType = abi.StaticBytesTypeSpec(length)
        assert staticBytesType == staticBytesType
        assert staticBytesType != abi.StaticBytesTypeSpec(length + 1)
        assert staticBytesType != abi.StaticArrayTypeSpec(
            abi.TupleTypeSpec(abi.ByteTypeSpec()), length
        )


def test_StaticArrayTypeSpec_is_dynamic():
    for elementType in STATIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            assert not staticArrayType.is_dynamic()

    for length in range(256):
        assert not abi.StaticBytesTypeSpec(length).is_dynamic()

    for elementType in DYNAMIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            assert staticArrayType.is_dynamic()


def test_StaticArrayTypeSpec_byte_length_static():
    for elementType in STATIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            actual = staticArrayType.byte_length_static()

            if elementType == abi.BoolTypeSpec():
                expected = _bool_sequence_length(length)
            else:
                expected = elementType.byte_length_static() * length

            assert (
                actual == expected
            ), "failed with element type {} and length {}".format(elementType, length)

    for length in range(256):
        staticBytesType = abi.StaticBytesTypeSpec(length)
        actual = staticBytesType.byte_length_static()
        assert (
            actual == length
        ), f"failed with element type {staticBytesType.value_type_spec()} and length {length}"

    for elementType in DYNAMIC_TYPES:
        for length in range(256):
            staticArrayType = abi.StaticArrayTypeSpec(elementType, length)
            with pytest.raises(ValueError):
                staticArrayType.byte_length_static()


def test_StaticArray_decode():
    encoded = pt.Bytes("encoded")
    for start_index in (None, pt.Int(1)):
        for end_index in (None, pt.Int(2)):
            for length in (None, pt.Int(3)):
                value = abi.StaticArray(
                    abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 10)
                )

                if end_index is not None and length is not None:
                    with pytest.raises(pt.TealInputError):
                        value.decode(
                            encoded,
                            start_index=start_index,
                            end_index=end_index,
                            length=length,
                        )
                    continue

                expr = value.decode(
                    encoded, start_index=start_index, end_index=end_index, length=length
                )
                assert expr.type_of() == pt.TealType.none
                assert not expr.has_return()

                expectedExpr = value.stored_value.store(
                    substring_for_decoding(
                        encoded,
                        start_index=start_index,
                        end_index=end_index,
                        length=length,
                    )
                )
                expected, _ = expectedExpr.__teal__(options)
                expected.addIncoming()
                expected = pt.TealBlock.NormalizeBlocks(expected)

                actual, _ = expr.__teal__(options)
                actual.addIncoming()
                actual = pt.TealBlock.NormalizeBlocks(actual)

                with pt.TealComponent.Context.ignoreExprEquality():
                    assert actual == expected


def test_StaticArray_set_values():
    value = abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 10))

    with pytest.raises(pt.TealInputError):
        value.set([])

    with pytest.raises(pt.TealInputError):
        value.set([abi.Uint64()] * 9)

    with pytest.raises(pt.TealInputError):
        value.set([abi.Uint64()] * 11)

    with pytest.raises(pt.TealInputError):
        value.set([abi.Uint16()] * 10)

    with pytest.raises(pt.TealInputError):
        value.set([abi.Uint64()] * 9 + [abi.Uint16()])

    values = [abi.Uint64() for _ in range(10)]
    expr = value.set(values)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expectedExpr = value.stored_value.store(_encode_tuple(values))
    expected, _ = expectedExpr.__teal__(options)
    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_StaticArray_set_copy():
    value = abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 10))
    otherArray = abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 10))

    with pytest.raises(pt.TealInputError):
        value.set(abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 11)))

    with pytest.raises(pt.TealInputError):
        value.set(abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint8TypeSpec(), 10)))

    with pytest.raises(pt.TealInputError):
        value.set(abi.Uint64())

    expr = value.set(otherArray)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.load, otherArray.stored_value.slot),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_StaticArray_set_computed():
    value = abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 10))
    computed = ContainerType(
        value.type_spec(), pt.Bytes("indeed this is hard to simulate")
    )
    expr = value.set(computed)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.byte, '"indeed this is hard to simulate"'),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )
    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = actual.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        value.set(
            ContainerType(
                abi.StaticArrayTypeSpec(abi.Uint16TypeSpec(), 40),
                pt.Bytes("well i am trolling"),
            )
        )


# AACS key recovery
BYTE_HEX_TEST_CASE = "09f911029d74e35bd84156c5635688c0"

BYTES_SET_TESTCASES = [
    bytes.fromhex(BYTE_HEX_TEST_CASE),
    bytearray.fromhex(BYTE_HEX_TEST_CASE),
]


@pytest.mark.parametrize("test_case", BYTES_SET_TESTCASES)
def test_StaticBytes_set_py_bytes(test_case: bytes | bytearray):
    value: abi.StaticBytes = abi.StaticBytes(abi.StaticBytesTypeSpec(len(test_case)))

    expr = value.set(test_case)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = actual.NormalizeBlocks(actual)

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.byte, "0x" + BYTE_HEX_TEST_CASE),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
        ]
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        value.set(test_case[:-1])


@pytest.mark.parametrize("test_case", BYTES_SET_TESTCASES)
def test_StaticBytes_expr(test_case: bytes | bytearray):
    value: abi.StaticBytes = abi.StaticBytes(
        abi.StaticBytesTypeSpec(len(test_case) * 2)
    )
    set_expr = pt.Concat(pt.Bytes(test_case), pt.Bytes(test_case))

    expr = value.set(set_expr)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = actual.NormalizeBlocks(actual)

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.byte, "0x" + BYTE_HEX_TEST_CASE),
            pt.TealOp(None, pt.Op.byte, "0x" + BYTE_HEX_TEST_CASE),
            pt.TealOp(None, pt.Op.concat),
            pt.TealOp(None, pt.Op.store, value.stored_value.slot),
            pt.TealOp(None, pt.Op.int, 32),
            pt.TealOp(None, pt.Op.load, value.stored_value.slot),
            pt.TealOp(None, pt.Op.len),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.assert_),
        ]
    )

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_StaticArray_encode():
    value = abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), 10))
    expr = value.encode()
    assert expr.type_of() == pt.TealType.bytes
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [pt.TealOp(None, pt.Op.load, value.stored_value.slot)]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_StaticArray_length():
    for length in (0, 1, 2, 3, 1000):
        value = abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), length))
        expr = value.length()
        assert expr.type_of() == pt.TealType.uint64
        assert not expr.has_return()

        expected = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, length)])

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected


def test_StaticArray_getitem():
    for length in (0, 1, 2, 3, 1000):
        value = abi.StaticArray(abi.StaticArrayTypeSpec(abi.Uint64TypeSpec(), length))

        for index in range(length):
            # dynamic indexes
            indexExpr = pt.Int(index)
            element = value[indexExpr]
            assert type(element) is abi.ArrayElement
            assert element.array is value
            assert element.index is indexExpr

        for index in range(length):
            # static indexes
            element = value[index]
            assert type(element) is abi.ArrayElement
            assert element.array is value
            assert type(element.index) is pt.Int
            assert element.index.value == index

        with pytest.raises(pt.TealInputError):
            value[-1]

        with pytest.raises(pt.TealInputError):
            value[length]



import pytest

import pyteal as pt

from tests.compile_asserts import assert_new_v_old

# ---- TESTS FOR PyTEAL THAT PREDATE PASS-BY-REF - assert that changes to compiler don't affect the generated TEAL ---- #


@pt.Subroutine(pt.TealType.bytes)
def logcat(some_bytes, an_int):
    catted = pt.ScratchVar(pt.TealType.bytes)
    return pt.Seq(
        catted.store(pt.Concat(some_bytes, pt.Itob(an_int))),
        pt.Log(catted.load()),
        catted.load(),
    )


def sub_logcat():
    return pt.Seq(
        pt.Assert(logcat(pt.Bytes("hello"), pt.Int(42)) == pt.Bytes("hello42")),
        pt.Int(1),
    )


@pt.Subroutine(pt.TealType.uint64)
def slow_fibonacci(n):
    return (
        pt.If(n <= pt.Int(1))
        .Then(n)
        .Else(slow_fibonacci(n - pt.Int(2)) + slow_fibonacci(n - pt.Int(1)))
    )


def sub_slowfib():
    return slow_fibonacci(pt.Int(3))


@pt.Subroutine(pt.TealType.uint64)
def fast_fibonacci(n):
    i = pt.ScratchVar(pt.TealType.uint64)
    a = pt.ScratchVar(pt.TealType.uint64)
    b = pt.ScratchVar(pt.TealType.uint64)
    return pt.Seq(
        a.store(pt.Int(0)),
        b.store(pt.Int(1)),
        pt.For(i.store(pt.Int(1)), i.load() <= n, i.store(i.load() + pt.Int(1))).Do(
            pt.Seq(
                b.store(a.load() + b.load()),
                a.store(b.load() - a.load()),
            )
        ),
        a.load(),
    )


def sub_fastfib():
    return fast_fibonacci(pt.Int(3))


@pt.Subroutine(pt.TealType.uint64)
def recursiveIsEven(i):
    return (
        pt.If(i == pt.Int(0))
        .Then(pt.Int(1))
        .ElseIf(i == pt.Int(1))
        .Then(pt.Int(0))
        .Else(recursiveIsEven(i - pt.Int(2)))
    )


def sub_even():
    return pt.Seq(
        pt.Pop(recursiveIsEven(pt.Int(1000))),
        recursiveIsEven(pt.Int(1001)),
    )


PT_CASES = (sub_logcat, sub_slowfib, sub_fastfib, sub_even)


@pytest.mark.parametrize("pt_case", PT_CASES)
def test_old(pt_case):
    assert_new_v_old(pt_case, 5, "pre_v6")



from enum import Enum

from pyteal.types import TealType, require_type
from pyteal.errors import verifyFieldVersion
from pyteal.ir import Op
from pyteal.ast.multi import MultiValue
from pyteal.ast.expr import Expr


class VrfVerifyStandard(Enum):
    # fmt: off
    #           id  |   name   | min version
    algorand =  (0, "VrfAlgorand",  7)  # noqa: E222
    chainlink = (1, "VrfChainlink", 7)
    # fmt: on

    def __init__(self, id: int, name: str, min_version: int) -> None:
        self.id = id
        self.arg_name = name
        self.min_version = min_version


VrfVerifyStandard.__module__ = "pyteal"


class VrfVerify(MultiValue):
    """An expression that verifies the proof of a message against a public key."""

    def __init__(
        self, standard: VrfVerifyStandard, message: Expr, proof: Expr, public_key: Expr
    ) -> None:
        require_type(message, TealType.bytes)
        require_type(proof, TealType.bytes)
        require_type(public_key, TealType.bytes)

        self.standard = standard

        super().__init__(
            Op.vrf_verify,
            [TealType.bytes, TealType.uint64],
            immediate_args=[standard.arg_name],
            args=[message, proof, public_key],
            compile_check=lambda options: verifyFieldVersion(
                standard.arg_name, standard.min_version, options.version
            ),
        )

    def __str__(self):
        return "(VrfVerify {})".format(self.standard.arg_name)

    @classmethod
    def algorand(cls, message: Expr, proof: Expr, public_key: Expr) -> "VrfVerify":
        """Verifies the proof of a message against a public key using the Algorand VRF standard.

        Args:
            message: The message to verify.
            proof: The proof of the message.
            public_key: The public key to use to verify the proof.

        Returns:
            A MultiValue expression representing the VRF output and a verification flag.
        """
        return cls(VrfVerifyStandard.algorand, message, proof, public_key)

    @classmethod
    def chainlink(cls, message: Expr, proof: Expr, public_key: Expr) -> "VrfVerify":
        """Verifies the proof of a message against a public key using the Chainlink VRF standard.

        Args:
            message: The message to verify.
            proof: The proof of the message.
            public_key: The public key to use to verify the proof.

        Returns:
            A MultiValue expression representing the VRF output and a verification flag.
        """
        return cls(VrfVerifyStandard.chainlink, message, proof, public_key)


VrfVerify.__module__ = "pyteal"



from typing import TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.errors import verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class UnaryExpr(Expr):
    """An expression with a single argument."""

    def __init__(
        self, op: Op, inputType: TealType, outputType: TealType, arg: Expr
    ) -> None:
        super().__init__()
        require_type(arg, inputType)
        self.op = op
        self.outputType = outputType
        self.arg = arg

    def __teal__(self, options: "CompileOptions"):
        verifyProgramVersion(
            self.op.min_version,
            options.version,
            "Program version too low to use op {}".format(self.op),
        )

        return TealBlock.FromOp(options, TealOp(self, self.op), self.arg)

    def __str__(self):
        return "({} {})".format(str(self.op).title().replace("_", ""), self.arg)

    def type_of(self):
        return self.outputType

    def has_return(self):
        return False


UnaryExpr.__module__ = "pyteal"


def Btoi(arg: Expr) -> UnaryExpr:
    """Convert a byte string to a uint64."""
    return UnaryExpr(Op.btoi, TealType.bytes, TealType.uint64, arg)


def Itob(arg: Expr) -> UnaryExpr:
    """Convert a uint64 string to a byte string."""
    return UnaryExpr(Op.itob, TealType.uint64, TealType.bytes, arg)


def Len(arg: Expr) -> UnaryExpr:
    """Get the length of a byte string."""
    return UnaryExpr(Op.len, TealType.bytes, TealType.uint64, arg)


def BitLen(arg: Expr) -> UnaryExpr:
    """Get the index of the highest nonzero bit in an integer.

    If the argument is 0, 0 will be returned.

    If the argument is a byte array, it is interpreted as a big-endian unsigned integer.

    Requires program version 4 or higher.
    """
    return UnaryExpr(Op.bitlen, TealType.anytype, TealType.uint64, arg)


def Sha256(arg: Expr) -> UnaryExpr:
    """Get the SHA-256 hash of a byte string."""
    return UnaryExpr(Op.sha256, TealType.bytes, TealType.bytes, arg)


def Sha512_256(arg: Expr) -> UnaryExpr:
    """Get the SHA-512/256 hash of a byte string."""
    return UnaryExpr(Op.sha512_256, TealType.bytes, TealType.bytes, arg)


def Sha3_256(arg: Expr) -> UnaryExpr:
    """Get the SHA3-256 hash of a byte string."""
    return UnaryExpr(Op.sha3_256, TealType.bytes, TealType.bytes, arg)


def Keccak256(arg: Expr) -> UnaryExpr:
    """Get the KECCAK-256 hash of a byte string."""
    return UnaryExpr(Op.keccak256, TealType.bytes, TealType.bytes, arg)


def Not(arg: Expr) -> UnaryExpr:
    """Get the logical inverse of a uint64.

    If the argument is 0, then this will produce 1. Otherwise this will produce 0.
    """
    return UnaryExpr(Op.logic_not, TealType.uint64, TealType.uint64, arg)


def BitwiseNot(arg: Expr) -> UnaryExpr:
    """Get the bitwise inverse of a uint64.

    Produces ~arg.
    """
    return UnaryExpr(Op.bitwise_not, TealType.uint64, TealType.uint64, arg)


def Sqrt(arg: Expr) -> UnaryExpr:
    """Get the integer square root of a uint64.

    This will return the largest integer X such that X^2 <= arg.

    Requires program version 4 or higher.
    """
    return UnaryExpr(Op.sqrt, TealType.uint64, TealType.uint64, arg)


def Pop(arg: Expr) -> UnaryExpr:
    """Pop a value from the stack."""
    return UnaryExpr(Op.pop, TealType.anytype, TealType.none, arg)


def Balance(account: Expr) -> UnaryExpr:
    """Get the balance of a user in microAlgos.

    Argument must be an index into Txn.Accounts that corresponds to the account to read from,
    must be evaluated to uint64 (or, since v4, an account address that appears in Txn.Accounts
    or is Txn.Sender).

    This operation is only permitted in application mode.
    """
    return UnaryExpr(Op.balance, TealType.anytype, TealType.uint64, account)


def MinBalance(account: Expr) -> UnaryExpr:
    """Get the minimum balance of a user in microAlgos.

    For more information about minimum balances, see: https://developer.algorand.org/docs/features/accounts/#minimum-balance

    Argument must be an index into Txn.Accounts that corresponds to the account to read from,
    must be evaluated to uint64 (or, since v4, an account address that appears in Txn.Accounts
    or is Txn.Sender).

    Requires program version 3 or higher. This operation is only permitted in application mode.
    """
    return UnaryExpr(Op.min_balance, TealType.anytype, TealType.uint64, account)


def BytesNot(arg: Expr) -> UnaryExpr:
    """Get the bitwise inverse of bytes.

    Produces ~arg.
    Argument must not exceed 64 bytes.

    Requires program version 4 or higher.
    """
    return UnaryExpr(Op.b_not, TealType.bytes, TealType.bytes, arg)


def BytesSqrt(arg: Expr) -> UnaryExpr:
    """Get the bytes square root of bytes.

    This will return the largest integer X such that X^2 <= arg.

    Requires program version 6 or higher.
    """
    return UnaryExpr(Op.bsqrt, TealType.bytes, TealType.bytes, arg)


def BytesZero(arg: Expr) -> UnaryExpr:
    """Get a byte-array of a specified length, containing all zero bytes.

    Argument must evaluate to uint64.

    Requires program version 4 or higher.
    """
    return UnaryExpr(Op.bzero, TealType.uint64, TealType.bytes, arg)


def Log(message: Expr) -> UnaryExpr:
    """Write a message to log state of the current application.

    This will fail if called more than :code:`MaxLogCalls` times in a program (32 as of AVM v5), or
    if the sum of the lengths of all logged messages in a program exceeds 1024 bytes.

    Args:
        message: The message to write. Must evaluate to bytes.

    Requires program version 5 or higher.
    """
    return UnaryExpr(Op.log, TealType.bytes, TealType.none, message)



from typing import List, Tuple, NamedTuple, Callable, Union, Optional
from pyteal.ast.abi.type_test import ContainerType
import pyteal as pt
from pyteal import abi

import pytest

options = pt.CompileOptions(version=5)


class UintTestData(NamedTuple):
    uintType: abi.UintTypeSpec
    instanceType: type
    expectedBits: int
    maxValue: int
    checkUpperBound: bool
    expectedDecoding: Callable[
        [pt.Expr, Optional[pt.Expr], Optional[pt.Expr], Optional[pt.Expr]], pt.Expr
    ]
    expectedEncoding: Callable[[abi.Uint], pt.Expr]


def noneToInt0(value: Union[None, pt.Expr]):
    if value is None:
        return pt.Int(0)
    return value


testData = [
    UintTestData(
        uintType=abi.Uint8TypeSpec(),
        instanceType=abi.Uint8,
        expectedBits=8,
        maxValue=2**8 - 1,
        checkUpperBound=True,
        expectedDecoding=lambda encoded, start_index, end_index, length: pt.GetByte(
            encoded, noneToInt0(start_index)
        ),
        expectedEncoding=lambda uintType: pt.SetByte(
            pt.Bytes(b"\x00"), pt.Int(0), uintType.get()
        ),
    ),
    UintTestData(
        uintType=abi.Uint16TypeSpec(),
        instanceType=abi.Uint16,
        expectedBits=16,
        maxValue=2**16 - 1,
        checkUpperBound=True,
        expectedDecoding=lambda encoded, start_index, end_index, length: pt.ExtractUint16(
            encoded, noneToInt0(start_index)
        ),
        expectedEncoding=lambda uintType: pt.Suffix(pt.Itob(uintType.get()), pt.Int(6)),
    ),
    UintTestData(
        uintType=abi.Uint32TypeSpec(),
        instanceType=abi.Uint32,
        expectedBits=32,
        maxValue=2**32 - 1,
        checkUpperBound=True,
        expectedDecoding=lambda encoded, start_index, end_index, length: pt.ExtractUint32(
            encoded, noneToInt0(start_index)
        ),
        expectedEncoding=lambda uintType: pt.Suffix(pt.Itob(uintType.get()), pt.Int(4)),
    ),
    UintTestData(
        uintType=abi.Uint64TypeSpec(),
        instanceType=abi.Uint64,
        expectedBits=64,
        maxValue=2**64 - 1,
        checkUpperBound=False,
        expectedDecoding=lambda encoded, start_index, end_index, length: pt.Btoi(
            encoded
        )
        if start_index is None and end_index is None and length is None
        else pt.ExtractUint64(encoded, noneToInt0(start_index)),
        expectedEncoding=lambda uintType: pt.Itob(uintType.get()),
    ),
]


def test_UintTypeSpec_bits():
    for test in testData:
        assert test.uintType.bit_size() == test.expectedBits
        assert test.uintType.byte_length_static() * 8 == test.expectedBits


def test_UintTypeSpec_str():
    for test in testData:
        assert str(test.uintType) == "uint{}".format(test.expectedBits)
    assert str(abi.ByteTypeSpec()) == "byte"


def test_UintTypeSpec_is_dynamic():
    for test in testData:
        assert not test.uintType.is_dynamic()
    assert not abi.ByteTypeSpec().is_dynamic()


def test_UintTypeSpec_eq():
    for i, test in enumerate(testData):
        assert test.uintType == test.uintType

        for j, otherTest in enumerate(testData):
            if i == j:
                continue
            assert test.uintType != otherTest.uintType

        for otherType in (
            abi.BoolTypeSpec(),
            abi.StaticArrayTypeSpec(test.uintType, 1),
            abi.DynamicArrayTypeSpec(test.uintType),
        ):
            assert test.uintType != otherType

    assert abi.ByteTypeSpec() != abi.Uint8TypeSpec()
    assert abi.Uint8TypeSpec() != abi.ByteTypeSpec()


def test_UintTypeSpec_storage_type():
    for test in testData:
        assert test.uintType.storage_type() == pt.TealType.uint64
    assert abi.BoolTypeSpec().storage_type() == pt.TealType.uint64


def test_UintTypeSpec_new_instance():
    for test in testData:
        assert isinstance(test.uintType.new_instance(), test.instanceType)
    assert isinstance(abi.ByteTypeSpec().new_instance(), abi.Byte)


def test_Uint_set_static():
    for test in testData:
        for value_to_set in (0, 1, 100, test.maxValue):
            value = test.uintType.new_instance()
            expr = value.set(value_to_set)
            assert expr.type_of() == pt.TealType.none
            assert not expr.has_return()

            expected = pt.TealSimpleBlock(
                [
                    pt.TealOp(None, pt.Op.int, value_to_set),
                    pt.TealOp(None, pt.Op.store, value.stored_value.slot),
                ]
            )

            actual, _ = expr.__teal__(options)
            actual.addIncoming()
            actual = pt.TealBlock.NormalizeBlocks(actual)

            with pt.TealComponent.Context.ignoreExprEquality():
                assert actual == expected

        with pytest.raises(pt.TealInputError):
            value.set(test.maxValue + 1)

        with pytest.raises(pt.TealInputError):
            value.set(-1)


def test_Uint_set_expr():
    for test in testData:
        value = test.uintType.new_instance()
        expr = value.set(pt.Int(10) + pt.Int(1))
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        upperBoundCheck = []
        if test.checkUpperBound:
            upperBoundCheck = [
                pt.TealOp(None, pt.Op.load, value.stored_value.slot),
                pt.TealOp(None, pt.Op.int, test.maxValue + 1),
                pt.TealOp(None, pt.Op.lt),
                pt.TealOp(None, pt.Op.assert_),
            ]

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(None, pt.Op.int, 10),
                pt.TealOp(None, pt.Op.int, 1),
                pt.TealOp(None, pt.Op.add),
                pt.TealOp(None, pt.Op.store, value.stored_value.slot),
            ]
            + upperBoundCheck
        )

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected


def test_Uint_set_copy():
    for test in testData:
        value = test.uintType.new_instance()
        other = test.uintType.new_instance()
        expr = value.set(other)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(None, pt.Op.load, other.stored_value.slot),
                pt.TealOp(None, pt.Op.store, value.stored_value.slot),
            ]
        )

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected

        with pytest.raises(pt.TealInputError):
            value.set(abi.Bool())


def test_Uint_set_computed():
    byte_computed_value = ContainerType(abi.ByteTypeSpec(), pt.Int(0x22))

    for test in testData:
        computed_value = ContainerType(test.uintType, pt.Int(0x44))
        value = test.uintType.new_instance()
        expr = value.set(computed_value)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(None, pt.Op.int, 0x44),
                pt.TealOp(None, pt.Op.store, value.stored_value.slot),
            ]
        )

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected

        with pytest.raises(pt.TealInputError):
            value.set(byte_computed_value)


def test_Uint_get():
    for test in testData:
        value = test.uintType.new_instance()
        expr = value.get()
        assert expr.type_of() == pt.TealType.uint64
        assert not expr.has_return()

        expected = pt.TealSimpleBlock(
            [pt.TealOp(expr, pt.Op.load, value.stored_value.slot)]
        )

        actual, _ = expr.__teal__(options)

        assert actual == expected


def test_Uint_decode():
    encoded = pt.Bytes("encoded")
    for test in testData:
        for start_index in (None, pt.Int(1)):
            for end_index in (None, pt.Int(2)):
                for length in (None, pt.Int(3)):
                    value = test.uintType.new_instance()
                    expr = value.decode(
                        encoded,
                        start_index=start_index,
                        end_index=end_index,
                        length=length,
                    )
                    assert expr.type_of() == pt.TealType.none
                    assert not expr.has_return()

                    expectedDecoding = value.stored_value.store(
                        test.expectedDecoding(encoded, start_index, end_index, length)
                    )
                    expected, _ = expectedDecoding.__teal__(options)
                    expected.addIncoming()
                    expected = pt.TealBlock.NormalizeBlocks(expected)

                    actual, _ = expr.__teal__(options)
                    actual.addIncoming()
                    actual = pt.TealBlock.NormalizeBlocks(actual)

                    with pt.TealComponent.Context.ignoreExprEquality():
                        assert actual == expected


def test_Uint_encode():
    for test in testData:
        value = test.uintType.new_instance()
        expr = value.encode()
        assert expr.type_of() == pt.TealType.bytes
        assert not expr.has_return()

        expected, _ = test.expectedEncoding(value).__teal__(options)
        expected.addIncoming()
        expected = pt.TealBlock.NormalizeBlocks(expected)

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected


def test_ByteUint8_mutual_conversion():
    cases: List[Tuple[abi.UintTypeSpec, abi.UintTypeSpec]] = [
        (abi.Uint8TypeSpec(), abi.ByteTypeSpec()),
        (abi.ByteTypeSpec(), abi.Uint8TypeSpec()),
    ]
    for type_a, type_b in cases:
        type_b_instance = type_b.new_instance()
        other = type_a.new_instance()
        expr = type_b_instance.set(other)

        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(None, pt.Op.load, other.stored_value.slot),
                pt.TealOp(None, pt.Op.store, type_b_instance.stored_value.slot),
            ]
        )

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_if_has_return():
    exprWithReturn = pt.If(pt.Int(1), pt.Return(pt.Int(1)), pt.Return(pt.Int(0)))
    assert exprWithReturn.has_return()

    exprWithoutReturn = pt.If(pt.Int(1), pt.Int(1), pt.Int(0))
    assert not exprWithoutReturn.has_return()

    exprSemiReturn = pt.If(
        pt.Int(1),
        pt.Return(pt.Int(1)),
        pt.App.globalPut(pt.Bytes("key"), pt.Bytes("value")),
    )
    assert not exprSemiReturn.has_return()


def test_if_int():
    args = [pt.Int(0), pt.Int(1), pt.Int(2)]
    expr = pt.If(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.uint64

    expected, _ = args[0].__teal__(options)
    thenBlock, _ = args[1].__teal__(options)
    elseBlock, _ = args[2].__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlock)
    expectedBranch.setFalseBlock(elseBlock)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlock.setNextBlock(end)
    elseBlock.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_if_bytes():
    args = [pt.Int(1), pt.Txn.sender(), pt.Txn.receiver()]
    expr = pt.If(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected, _ = args[0].__teal__(options)
    thenBlock, _ = args[1].__teal__(options)
    elseBlock, _ = args[2].__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlock)
    expectedBranch.setFalseBlock(elseBlock)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlock.setNextBlock(end)
    elseBlock.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_if_none():
    args = [pt.Int(0), pt.Pop(pt.Txn.sender()), pt.Pop(pt.Txn.receiver())]
    expr = pt.If(args[0], args[1], args[2])
    assert expr.type_of() == pt.TealType.none

    expected, _ = args[0].__teal__(options)
    thenBlockStart, thenBlockEnd = args[1].__teal__(options)
    elseBlockStart, elseBlockEnd = args[2].__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlockStart)
    expectedBranch.setFalseBlock(elseBlockStart)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlockEnd.setNextBlock(end)
    elseBlockEnd.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_if_single():
    args = [pt.Int(1), pt.Pop(pt.Int(1))]
    expr = pt.If(args[0], args[1])
    assert expr.type_of() == pt.TealType.none

    expected, _ = args[0].__teal__(options)
    thenBlockStart, thenBlockEnd = args[1].__teal__(options)
    end = pt.TealSimpleBlock([])
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlockStart)
    expectedBranch.setFalseBlock(end)
    expected.setNextBlock(expectedBranch)
    thenBlockEnd.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_if_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.If(pt.Int(0), pt.Txn.amount(), pt.Txn.sender())

    with pytest.raises(pt.TealTypeError):
        pt.If(pt.Txn.sender(), pt.Int(1), pt.Int(0))

    with pytest.raises(pt.TealTypeError):
        pt.If(pt.Int(0), pt.Txn.sender())

    with pytest.raises(pt.TealTypeError):
        pt.If(pt.Int(0), pt.Int(2))

    with pytest.raises(pt.TealCompileError):
        expr = pt.If(pt.Int(0))
        expr.__teal__(options)


def test_if_alt_int():
    args = [pt.Int(0), pt.Int(1), pt.Int(2)]
    expr = pt.If(args[0]).Then(args[1]).Else(args[2])
    assert expr.type_of() == pt.TealType.uint64

    expected, _ = args[0].__teal__(options)
    thenBlock, _ = args[1].__teal__(options)
    elseBlock, _ = args[2].__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlock)
    expectedBranch.setFalseBlock(elseBlock)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlock.setNextBlock(end)
    elseBlock.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_if_alt_bytes():
    args = [pt.Int(1), pt.Txn.sender(), pt.Txn.receiver()]
    expr = pt.If(args[0]).Then(args[1]).Else(args[2])
    assert expr.type_of() == pt.TealType.bytes

    expected, _ = args[0].__teal__(options)
    thenBlock, _ = args[1].__teal__(options)
    elseBlock, _ = args[2].__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlock)
    expectedBranch.setFalseBlock(elseBlock)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlock.setNextBlock(end)
    elseBlock.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_if_alt_none():
    args = [pt.Int(0), pt.Pop(pt.Txn.sender()), pt.Pop(pt.Txn.receiver())]
    expr = pt.If(args[0]).Then(args[1]).Else(args[2])
    assert expr.type_of() == pt.TealType.none

    expected, _ = args[0].__teal__(options)
    thenBlockStart, thenBlockEnd = args[1].__teal__(options)
    elseBlockStart, elseBlockEnd = args[2].__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlockStart)
    expectedBranch.setFalseBlock(elseBlockStart)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlockEnd.setNextBlock(end)
    elseBlockEnd.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_elseif_syntax():
    args = [pt.Int(0), pt.Int(1), pt.Int(2), pt.Int(3), pt.Int(4)]
    expr = pt.If(args[0]).Then(args[1]).ElseIf(args[2]).Then(args[3]).Else(args[4])
    assert expr.type_of() == pt.TealType.uint64

    elseExpr = pt.If(args[2]).Then(args[3]).Else(args[4])
    expected, _ = args[0].__teal__(options)
    thenBlock, _ = args[1].__teal__(options)
    elseStart, elseEnd = elseExpr.__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlock)
    expectedBranch.setFalseBlock(elseStart)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlock.setNextBlock(end)
    elseEnd.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_elseif_multiple():
    args = [pt.Int(0), pt.Int(1), pt.Int(2), pt.Int(3), pt.Int(4), pt.Int(5), pt.Int(6)]
    expr = (
        pt.If(args[0])
        .Then(args[1])
        .ElseIf(args[2])
        .Then(args[3])
        .ElseIf(args[4])
        .Then(args[5])
        .Else(args[6])
    )
    assert expr.type_of() == pt.TealType.uint64

    elseIfExpr = pt.If(args[2], args[3], pt.If(args[4], args[5], args[6]))
    expected, _ = args[0].__teal__(options)
    thenBlock, _ = args[1].__teal__(options)
    elseStart, elseEnd = elseIfExpr.__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlock)
    expectedBranch.setFalseBlock(elseStart)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlock.setNextBlock(end)
    elseEnd.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_if_invalid_alt_syntax():
    with pytest.raises(pt.TealCompileError):
        expr = pt.If(pt.Int(0)).ElseIf(pt.Int(1))
        expr.__teal__(options)

    with pytest.raises(pt.TealCompileError):
        expr = pt.If(pt.Int(0)).ElseIf(pt.Int(1)).Then(pt.Int(2))
        expr.__teal__(options)

    with pytest.raises(pt.TealCompileError):
        expr = pt.If(pt.Int(0)).Then(pt.Int(1)).ElseIf(pt.Int(2))
        expr.__teal__(options)

    with pytest.raises(pt.TealCompileError):
        expr = pt.If(pt.Int(0)).Then(pt.Int(1)).ElseIf(pt.Int(2))
        expr.__teal__(options)

    with pytest.raises(pt.TealTypeError):
        expr = pt.If(pt.Int(0)).Then(pt.Int(2))
        expr.type_of()

    with pytest.raises(pt.TealTypeError):
        expr = pt.If(pt.Int(0)).Then(pt.Pop(pt.Int(1)), pt.Int(2))
        expr.type_of()

    with pytest.raises(pt.TealInputError):
        pt.If(pt.Int(0)).Else(pt.Int(1)).Then(pt.Int(2))

    with pytest.raises(pt.TealInputError):
        expr = pt.If(pt.Int(0)).Else(pt.Int(1))
        expr.__teal__(options)

    with pytest.raises(pt.TealInputError):
        expr = pt.If(pt.Int(0)).Else(pt.Int(1)).Then(pt.Int(2))

    with pytest.raises(pt.TealInputError):
        expr = pt.If(pt.Int(0)).Else(pt.Int(1)).Else(pt.Int(2))

    with pytest.raises(pt.TealInputError):
        expr = pt.If(pt.Int(0), pt.Pop(pt.Int(1))).Else(pt.Int(2))


def test_if_alt_multi():
    args = [pt.Int(0), [pt.Pop(pt.Int(1)), pt.Int(2)], pt.Int(3)]
    expr = pt.If(args[0]).Then(*args[1]).Else(args[2])

    expected, _ = args[0].__teal__(options)
    thenBlockStart, thenBlockEnd = pt.Seq(*args[1]).__teal__(options)
    elseBlockStart, elseBlockEnd = args[2].__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlockStart)
    expectedBranch.setFalseBlock(elseBlockStart)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlockEnd.setNextBlock(end)
    elseBlockEnd.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_else_alt_multi():
    args = [pt.Int(0), pt.Int(1), [pt.Pop(pt.Int(2)), pt.Int(3)]]
    expr = pt.If(args[0]).Then(args[1]).Else(*args[2])

    expected, _ = args[0].__teal__(options)
    thenBlockStart, thenBlockEnd = args[1].__teal__(options)
    elseBlockStart, elseBlockEnd = pt.Seq(*args[2]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlockStart)
    expectedBranch.setFalseBlock(elseBlockStart)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlockEnd.setNextBlock(end)
    elseBlockEnd.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_elseif_multiple_with_multi():
    args = [
        pt.Int(0),
        [pt.Pop(pt.Int(1)), pt.Int(2)],
        pt.Int(3),
        [pt.Pop(pt.Int(4)), pt.Int(5)],
        pt.Int(6),
        [pt.Pop(pt.Int(7)), pt.Int(8)],
        [pt.Pop(pt.Int(9)), pt.Int(10)],
    ]
    expr = (
        pt.If(args[0])
        .Then(*args[1])
        .ElseIf(args[2])
        .Then(*args[3])
        .ElseIf(args[4])
        .Then(*args[5])
        .Else(*args[6])
    )

    elseIfExpr = pt.If(
        args[2], pt.Seq(args[3]), pt.If(args[4], pt.Seq(args[5]), pt.Seq(args[6]))
    )
    expected, _ = args[0].__teal__(options)
    thenBlock, thenBlockEnd = pt.Seq(args[1]).__teal__(options)
    elseStart, elseBlockEnd = elseIfExpr.__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(thenBlock)
    expectedBranch.setFalseBlock(elseStart)
    expected.setNextBlock(expectedBranch)
    end = pt.TealSimpleBlock([])
    thenBlockEnd.setNextBlock(end)
    elseBlockEnd.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected



from typing import Union, Sequence, cast
from collections.abc import Sequence as CollectionSequence

from algosdk.abi import ABIType

from pyteal.ast.abi.uint import Byte
from pyteal.ast.abi.type import ComputedValue, BaseType
from pyteal.ast.abi.array_dynamic import DynamicArray, DynamicArrayTypeSpec
from pyteal.ast.abi.uint import ByteTypeSpec, Uint16TypeSpec

from pyteal.ast.int import Int
from pyteal.ast.expr import Expr
from pyteal.ast.bytes import Bytes
from pyteal.ast.seq import Seq
from pyteal.ast.scratchvar import ScratchVar
from pyteal.ast.unaryexpr import Itob, Len
from pyteal.ast.substring import Suffix
from pyteal.ast.naryexpr import Concat

from pyteal.errors import TealInputError


def _encoded_byte_string(s: bytes | bytearray) -> Expr:
    prefix = ABIType.from_string("uint16").encode(len(s))
    return Bytes(prefix + s)


def _store_encoded_expr_byte_string_into_var(value: Expr, location: ScratchVar) -> Expr:
    return Seq(
        location.store(value),
        location.store(
            Concat(Suffix(Itob(Len(location.load())), Int(6)), location.load())
        ),
    )


class StringTypeSpec(DynamicArrayTypeSpec):
    def __init__(self) -> None:
        super().__init__(ByteTypeSpec())

    def new_instance(self) -> "String":
        return String()

    def annotation_type(self) -> "type[String]":
        return String

    def __str__(self) -> str:
        return "string"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, StringTypeSpec)


StringTypeSpec.__module__ = "pyteal.abi"


class String(DynamicArray[Byte]):
    def __init__(self) -> None:
        super().__init__(StringTypeSpec())

    def type_spec(self) -> StringTypeSpec:
        return StringTypeSpec()

    def get(self) -> Expr:
        """Return the value held by this String as a PyTeal expression.

        The expression will have the type TealType.bytes.
        """
        return Suffix(
            self.stored_value.load(), Int(Uint16TypeSpec().byte_length_static())
        )

    def set(
        self,
        value: Union[
            str,
            bytes,
            Expr,
            Sequence[Byte],
            DynamicArray[Byte],
            ComputedValue[DynamicArray[Byte]],
            "String",
            ComputedValue["String"],
        ],
    ) -> Expr:
        """Set the value of this String to the input value.

        The behavior of this method depends on the input argument type:

            * :code:`str`: set the value to the Python string.
            * :code:`bytes`: set the value to the Python byte string.
            * :code:`Expr`: set the value to the result of a PyTeal expression, which must evaluate to a TealType.bytes.
            * :code:`Sequence[Byte]`: set the bytes of this String to those contained in this Python sequence (e.g. a list or tuple).
            * :code:`DynamicArray[Byte]`: copy the bytes from a DynamicArray.
            * :code:`ComputedValue[DynamicArray[Byte]]`: copy the bytes from a DynamicArray produced by a ComputedValue.
            * :code:`String`: copy the value from another String.
            * :code:`ComputedValue[String]`: copy the value from a String produced by a ComputedValue.

        Args:
            value: The new value this String should take. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this String.
        """

        match value:
            case ComputedValue():
                return self._set_with_computed_type(value)
            case BaseType():
                if value.type_spec() == StringTypeSpec() or (
                    value.type_spec() == DynamicArrayTypeSpec(ByteTypeSpec())
                ):
                    return self.stored_value.store(value.stored_value.load())

                raise TealInputError(
                    f"Got {value} with type spec {value.type_spec()}, expected {StringTypeSpec}"
                )
            case bytes() | bytearray():
                return self.stored_value.store(_encoded_byte_string(value))
            case str():
                return self.stored_value.store(_encoded_byte_string(value.encode()))
            case Expr():
                return _store_encoded_expr_byte_string_into_var(
                    value, self.stored_value
                )
            case CollectionSequence():
                return super().set(cast(Sequence[Byte], value))

        raise TealInputError(
            f"Got {type(value)}, expected DynamicArray, ComputedValue, String, str, bytes, Expr"
        )


String.__module__ = "pyteal.abi"



from typing import (
    Union,
    Optional,
    Final,
    cast,
)
from abc import abstractmethod

from pyteal.types import TealType
from pyteal.errors import TealInputError
from pyteal.ast.scratchvar import ScratchVar
from pyteal.ast.expr import Expr
from pyteal.ast.seq import Seq
from pyteal.ast.assert_ import Assert
from pyteal.ast.substring import Suffix
from pyteal.ast.int import Int
from pyteal.ast.bytes import Bytes
from pyteal.ast.unaryexpr import Itob, Btoi
from pyteal.ast.binaryexpr import GetByte, ExtractUint16, ExtractUint32, ExtractUint64
from pyteal.ast.ternaryexpr import SetByte
from pyteal.ast.abi.type import ComputedValue, TypeSpec, BaseType

NUM_BITS_IN_BYTE = 8

SUPPORTED_UINT_SIZES = (8, 16, 32, 64)


def uint_storage_type(size: int) -> TealType:
    if size <= 64:
        return TealType.uint64
    return TealType.bytes


def uint_set(size: int, uint_var: ScratchVar, value: Union[int, Expr, "Uint"]) -> Expr:
    if size > 64:
        raise NotImplementedError(
            "Uint operations have not yet been implemented for bit sizes larger than 64"
        )

    checked = False
    if type(value) is int:
        if value >= 2**size:
            raise TealInputError("Value exceeds uint{} maximum: {}".format(size, value))
        value = Int(value)
        checked = True

    if isinstance(value, Uint):
        value = value.get()
        checked = True

    if checked or size == 64:
        return uint_var.store(cast(Expr, value))

    return Seq(
        uint_var.store(cast(Expr, value)),
        Assert(uint_var.load() < Int(2**size)),
    )


def uint_decode(
    size: int,
    uint_var: ScratchVar,
    encoded: Expr,
    start_index: Optional[Expr],
    end_index: Optional[Expr],
    length: Optional[Expr],
) -> Expr:
    if size > 64:
        raise NotImplementedError(
            "Uint operations have not yet been implemented for bit sizes larger than 64"
        )

    if size == 64:
        if start_index is None:
            if end_index is None and length is None:
                return uint_var.store(Btoi(encoded))
            start_index = Int(0)
        return uint_var.store(ExtractUint64(encoded, start_index))

    if start_index is None:
        start_index = Int(0)

    if size == 8:
        return uint_var.store(GetByte(encoded, start_index))
    if size == 16:
        return uint_var.store(ExtractUint16(encoded, start_index))
    if size == 32:
        return uint_var.store(ExtractUint32(encoded, start_index))

    raise ValueError("Unsupported uint size: {}".format(size))


def uint_encode(size: int, uint_var: Expr | ScratchVar) -> Expr:

    if isinstance(uint_var, ScratchVar):
        uint_var = uint_var.load()

    if size > 64:
        raise NotImplementedError(
            "Uint operations have not yet been implemented for bit sizes larger than 64"
        )

    if size == 8:
        return SetByte(Bytes(b"\x00"), Int(0), uint_var)
    if size == 16:
        return Suffix(Itob(uint_var), Int(6))
    if size == 32:
        return Suffix(Itob(uint_var), Int(4))
    if size == 64:
        return Itob(uint_var)

    raise ValueError("Unsupported uint size: {}".format(size))


class UintTypeSpec(TypeSpec):
    def __init__(self, bit_size: int) -> None:
        super().__init__()
        if bit_size not in SUPPORTED_UINT_SIZES:
            raise TypeError("Unsupported uint size: {}".format(bit_size))
        self.size: Final = bit_size

    @abstractmethod
    def new_instance(self) -> "Uint":
        pass

    @abstractmethod
    def annotation_type(self) -> "type[Uint]":
        pass

    def bit_size(self) -> int:
        """Get the bit size of this uint type"""
        return self.size

    def is_dynamic(self) -> bool:
        return False

    def byte_length_static(self) -> int:
        return self.bit_size() // NUM_BITS_IN_BYTE

    def storage_type(self) -> TealType:
        return uint_storage_type(self.bit_size())

    def __eq__(self, other: object) -> bool:
        # NOTE: by this implementation, ByteTypeSpec() != Uint8TypeSpec()
        return (
            type(self) is type(other)
            and self.bit_size() == cast(UintTypeSpec, other).bit_size()
        )

    def __str__(self) -> str:
        return "uint{}".format(self.bit_size())


UintTypeSpec.__module__ = "pyteal.abi"


class ByteTypeSpec(UintTypeSpec):
    def __init__(self) -> None:
        super().__init__(8)

    def new_instance(self) -> "Byte":
        return Byte()

    def annotation_type(self) -> "type[Byte]":
        return Byte

    def __str__(self) -> str:
        return "byte"


ByteTypeSpec.__module__ = "pyteal.abi"


class Uint8TypeSpec(UintTypeSpec):
    def __init__(self) -> None:
        super().__init__(8)

    def new_instance(self) -> "Uint8":
        return Uint8()

    def annotation_type(self) -> "type[Uint8]":
        return Uint8


Uint8TypeSpec.__module__ = "pyteal.abi"


class Uint16TypeSpec(UintTypeSpec):
    def __init__(self) -> None:
        super().__init__(16)

    def new_instance(self) -> "Uint16":
        return Uint16()

    def annotation_type(self) -> "type[Uint16]":
        return Uint16


Uint16TypeSpec.__module__ = "pyteal.abi"


class Uint32TypeSpec(UintTypeSpec):
    def __init__(self) -> None:
        super().__init__(32)

    def new_instance(self) -> "Uint32":
        return Uint32()

    def annotation_type(self) -> "type[Uint32]":
        return Uint32


Uint32TypeSpec.__module__ = "pyteal.abi"


class Uint64TypeSpec(UintTypeSpec):
    def __init__(self) -> None:
        super().__init__(64)

    def new_instance(self) -> "Uint64":
        return Uint64()

    def annotation_type(self) -> "type[Uint64]":
        return Uint64


Uint64TypeSpec.__module__ = "pyteal.abi"


class Uint(BaseType):
    @abstractmethod
    def __init__(self, spec: UintTypeSpec) -> None:
        super().__init__(spec)

    def type_spec(self) -> UintTypeSpec:
        return cast(UintTypeSpec, super().type_spec())

    def get(self) -> Expr:
        """Return the value held by this Uint as a PyTeal expression.

        The expression will have the type TealType.uint64.
        """
        return self.stored_value.load()

    def set(self, value: Union[int, Expr, "Uint", ComputedValue["Uint"]]) -> Expr:
        """Set the value of this Uint to the input value.

        There are a variety of ways to express the input value. Regardless of the type used to
        indicate the input value, this Uint type can only hold values in the range :code:`[0,2^N)`,
        where :code:`N` is the bit size of this Uint.

        The behavior of this method depends on the input argument type:

            * :code:`int`: set the value to a Python integer. A compiler error will occur if this value overflows or underflows this integer type.
            * :code:`Expr`: set the value to the result of a PyTeal expression, which must evaluate to a TealType.uint64. The program will fail if the evaluated value overflows or underflows this integer type.
            * :code:`Uint`: copy the value from another Uint. The argument's type must exactly match this integer's type, otherwise an error will occur. For example, it's not possible to set a Uint64 to a Uint8, or vice versa.
            * :code:`ComputedValue[Uint]`: copy the value from a Uint produced by a ComputedValue. The type produced by the ComputedValue must exactly match this integer's type, otherwise an error will occur.

        Args:
            value: The new value this Uint should take. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this Uint.
        """
        if isinstance(value, ComputedValue):
            return self._set_with_computed_type(value)

        if isinstance(value, BaseType) and not (
            isinstance(value.type_spec(), UintTypeSpec)
            and self.type_spec().bit_size()
            == cast(UintTypeSpec, value.type_spec()).bit_size()
        ):
            raise TealInputError(
                "Type {} is not assignable to type {}".format(
                    value.type_spec(), self.type_spec()
                )
            )
        return uint_set(self.type_spec().bit_size(), self.stored_value, value)

    def decode(
        self,
        encoded: Expr,
        *,
        start_index: Expr = None,
        end_index: Expr = None,
        length: Expr = None,
    ) -> Expr:
        return uint_decode(
            self.type_spec().bit_size(),
            self.stored_value,
            encoded,
            start_index,
            end_index,
            length,
        )

    def encode(self) -> Expr:
        return uint_encode(self.type_spec().bit_size(), self.stored_value)


Uint.__module__ = "pyteal.abi"


class Byte(Uint):
    def __init__(self) -> None:
        super().__init__(ByteTypeSpec())


Byte.__module__ = "pyteal.abi"


class Uint8(Uint):
    def __init__(self) -> None:
        super().__init__(Uint8TypeSpec())


Uint8.__module__ = "pyteal.abi"


class Uint16(Uint):
    def __init__(self) -> None:
        super().__init__(Uint16TypeSpec())


Uint16.__module__ = "pyteal.abi"


class Uint32(Uint):
    def __init__(self) -> None:
        super().__init__(Uint32TypeSpec())


Uint32.__module__ = "pyteal.abi"


class Uint64(Uint):
    def __init__(self) -> None:
        super().__init__(Uint64TypeSpec())


Uint64.__module__ = "pyteal.abi"



from typing import Optional, TYPE_CHECKING

from pyteal.ir.tealcomponent import TealComponent
from pyteal.ir.labelref import LabelReference

if TYPE_CHECKING:
    from pyteal.ast import Expr


class TealLabel(TealComponent):
    def __init__(
        self, expr: Optional["Expr"], label: LabelReference, comment: str = None
    ) -> None:
        super().__init__(expr)
        self.label = label
        self.comment = comment

    def getLabelRef(self) -> LabelReference:
        return self.label

    def assemble(self) -> str:
        comment = "\n// {}\n".format(self.comment) if self.comment is not None else ""
        return "{}{}:".format(comment, self.label.getLabel())

    def __repr__(self) -> str:
        return "TealLabel({}, {}, {})".format(
            self.expr, repr(self.label), repr(self.comment)
        )

    def __hash__(self) -> int:
        return hash((self.label, self.comment))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TealLabel):
            return False
        if TealComponent.Context.checkExprEquality and self.expr is not other.expr:
            return False
        return self.label == other.label and self.comment == other.comment


TealLabel.__module__ = "pyteal"



from typing import Callable, Generic, Sequence, TypeVar, cast
from dataclasses import dataclass

import algosdk.abi
from algosdk.v2client import algod

from graviton import blackbox
from graviton.blackbox import DryRunInspector, DryRunExecutor

from pyteal.ast.subroutine import OutputKwArgInfo

from pyteal import (
    abi,
    Arg,
    Btoi,
    Bytes,
    compileTeal,
    Expr,
    Int,
    Itob,
    Len,
    Log,
    Mode,
    Pop,
    ScratchVar,
    Seq,
    SubroutineFnWrapper,
    TealType,
    Txn,
)

from pyteal.ast.subroutine import ABIReturnSubroutine

# ---- Clients ---- #


def algod_with_assertion():
    algod = _algod_client()
    assert algod.status(), "algod.status() did not produce any results"
    return algod


def _algod_client(
    algod_address="http://localhost:4001", algod_token="a" * 64
) -> algod.AlgodClient:
    """Instantiate and return Algod client object."""
    return algod.AlgodClient(algod_token, algod_address)


# ---- Decorator ---- #


class BlackboxWrapper:
    def __init__(
        self,
        subr: SubroutineFnWrapper | ABIReturnSubroutine,
        input_types: list[TealType | None],
    ):
        subr.subroutine._validate(input_types=input_types)
        self.subroutine: SubroutineFnWrapper | ABIReturnSubroutine = subr
        self.input_types: list[TealType | abi.TypeSpec | None] = self._fill(input_types)

    def __call__(self, *args: Expr | ScratchVar, **kwargs) -> Expr | abi.ReturnedValue:
        return self.subroutine(*args, **kwargs)

    def name(self) -> str:
        return self.subroutine.name()

    def _fill(
        self, input_types: list[TealType | None]
    ) -> list[TealType | abi.TypeSpec | None]:
        match self.subroutine:
            case SubroutineFnWrapper() | ABIReturnSubroutine():
                args = self.subroutine.subroutine.arguments()
                abis = self.subroutine.subroutine.abi_args
                return [(x if x else abis[args[i]]) for i, x in enumerate(input_types)]
            case _:
                raise AssertionError(
                    f"Cannot handle subroutine of type {type(self.subroutine)}"
                )


def Blackbox(input_types: list[TealType | None]):
    """
    Decorator for transforming @Subroutine and @ABIReturnSubroutine wrapped functions
    into PyTeal expressions that compile into executable Teal programs.

    input_types: list[TealType] (required)
        List shadowing the input arguments of the decorated subroutine. In particular:
            * the list needs to be the same length as the number of subroutine arguments
            * if the subroutine argument is an ABI type, the shadowing input_type must be None
                as it will be determined at compile time from the subroutine's annotation
            * if the subroutine argument is an Expr or a ScratchVar, the shadowing input_type
                must be a TealType of the same kind as expected in the argument

    Some _Correct_ Examples:

    @Blackbox(input_types=[TealType.bytes, TealType.uint64])
    @Subroutine(TealType.bytes)
    def string_mult(x, y):
        ...

    @Blackbox(input_types=[TealType.bytes, TealType.uint64])
    @Subroutine(TealType.bytes)
    def string_mult(x: Expr, y: Expr):
        ...

    @Blackbox(input_types=[TealType.bytes, TealType.uint64])
    @Subroutine(TealType.bytes)
    def string_mult(x: ScratchVar, y: Expr):
        ...


    @Blackbox(input_types=[None, None])
    @ABIReturnSubroutine
    def string_mult(x: abi.String, y: abi.Uint16):
        ...

    @Blackbox(input_types=[None, TealType.uint64])
    @ABIReturnSubroutine
    def string_mult(x: abi.String, y):
        ...

    @Blackbox([None])
    @Subroutine(TealType.uint64)
    def cubed(n: abi.Uint64):
        ...

    """

    def decorator_blackbox(func: SubroutineFnWrapper | ABIReturnSubroutine):
        return BlackboxWrapper(func, input_types)

    return decorator_blackbox


# ---- API ---- #


Output = TypeVar("Output")
Lazy = Callable[[], Output]


@dataclass(frozen=True)
class _MatchMode(Generic[Output]):
    app_case: Lazy
    signature_case: Lazy

    def __call__(self, mode: Mode, *args, **kwargs) -> Output:
        match mode:
            case Mode.Application:
                return self.app_case()
            case Mode.Signature:
                return self.signature_case()
            case _:
                raise Exception(f"Unknown mode {mode} of type {type(mode)}")


def mode_to_execution_mode(mode: Mode) -> blackbox.ExecutionMode:
    return _MatchMode(
        app_case=lambda: blackbox.ExecutionMode.Application,
        signature_case=lambda: blackbox.ExecutionMode.Signature,
    )(mode)


class PyTealDryRunExecutor:
    def __init__(self, subr: BlackboxWrapper, mode: Mode):
        """
        Args:
            subr: a Subroutine or ABIReturnSubroutine which has been decorated with @Blackbox.
                Note: the `input_types` parameters should be supplied to the @Blackbox() decorator
                    cf. the Blackbox class for futher details about acceptable `input_types`

            mode: type of program to produce: logic sig (Mode.Signature) or app (Mode.Application)
        """
        input_types = subr.input_types
        assert (
            input_types is not None
        ), "please provide input_types in your @Subroutine or @ABIReturnSubroutine annotation (this is crucial for generating proper end-to-end testable PyTeal)"

        self.subr, self.mode, self.input_types = subr, mode, input_types
        match subr.subroutine:
            case SubroutineFnWrapper():
                approval = self._handle_SubroutineFnWrapper()
            case ABIReturnSubroutine():
                approval = self._handle_ABIReturnSubroutine()
            case _:
                raise AssertionError(
                    f"Cannot produce Blackbox pyteal for provided subroutine of type {type(subr.subroutine)}"
                )

        self._pyteal_lambda: Callable[..., Expr] = approval

    def is_abi(self) -> bool:
        return isinstance(self.subr.subroutine, ABIReturnSubroutine)

    def abi_argument_types(self) -> None | list[algosdk.abi.ABIType]:
        if not (self.input_types or self.is_abi()):
            return None

        def handle_arg(arg):
            if isinstance(arg, abi.TypeSpec):
                return abi.algosdk_from_type_spec(arg)
            return None

        return [handle_arg(arg) for arg in self.input_types]

    def abi_return_type(self) -> None | algosdk.abi.ABIType:
        if not self.is_abi():
            return None

        out_info = getattr(self.subr.subroutine, "output_kwarg_info")
        if not out_info:
            return None

        return abi.algosdk_from_type_spec(cast(OutputKwArgInfo, out_info).abi_type)

    def program(self) -> Expr:
        """Get ready-to-compile PyTeal program from Subroutines and ABIReturnSubroutines

        Returns:
            a PyTeal expression representing a ready-to-run TEAL program

        Generated TEAL code depends on the self.subr's type, the mode, the input types, and output type
        * logic sigs:
            * input received via `arg i`
            * args are converted (cf. "input conversion" below) and passed to the subroutine
            * subroutine output is not logged (log is not available)
            * in the case of ABIReturnSubroutine: the output is encoded on to the stack an then popped off
            * subroutine output is converted (cf "output conversion" below)
        * apps:
            * input received via `txna ApplicationArgs i`
            * args are converted (cf. "input conversion" below) and passed to the subroutine
            * the output is logged in the following ways:
                * Subroutine: logged after possible conversion (cf. "logging conversion")
                * ABIReturnSubroutine: the encoded output is concatenated to the return method selector and then logged
            * subroutine output is converted (cf "output conversion" below) (Subroutine case only)
        * input conversion:
            * Empty input array:
                do not read any args and call subroutine immediately
            * Expr arg of TealType.bytes and TealType.anytype:
                read arg and pass to subroutine as is
            * Expr arg of TealType.uint64:
                convert arg to int using Btoi() when received
            * pass-by-ref ScratchVar arguments:
                in addition to the above -
                    o store the arg (or converted arg) in a ScratchVar
                    o invoke the subroutine using this ScratchVar instead of the arg (or converted arg)
            * ABI arguments:
                in addition to the above -
                    o store the decoded arg into the ScratchVar of an ABI Type instance
                    o invoke the subroutine using this ABI Type instead of the arg
        * output conversion:
            * Subroutine case:
                * TealType.uint64:
                    provide subroutine's result to the top of the stack when exiting program
                * TealType.bytes:
                    convert subroutine's result to the top of the stack to its length and then exit
                * TealType.none or TealType.anytype:
                    push Int(1337) to the stack as it is either impossible (TealType.none),
                    or unknown at compile time (TealType.anytype) to convert to an Int
            * ABIReturnSubroutine case:
                * when present, the output is encoded as TealType.bytes which can be decoded by the receiver using
                appropriate ABI-libraries
        * logging conversion:
            * Subroutine case:
                * TealType.uint64:
                    convert subroutine's output using Itob() and log the result
                * TealType.bytes:
                    log the subroutine's result
                * TealType.none or TealType.anytype:
                    log Itob(Int(1337)) as it is either impossible (TealType.none),
                    or unknown at compile time (TealType.anytype) how to convert to Bytes
            * ABIReturnSubroutine case:
                * when present, the output is encoded as TealType.bytes and concatenated to the rewturn
                method selector. This can be decoded by the receiver using appropriate ABI-libraries

        For illustrative examples of how to use this method please refer to the integration test file `graviton_test.py` and especially:

        * `blackbox_pyteal_example1()`: Using blackbox_pyteal() for a simple test of both an app and logic sig
        * `blackbox_pyteal_example2()`: Using blackbox_pyteal() to make 400 assertions and generate a CSV report with 400 dryrun rows
        * `blackbox_pyteal_example3()`: declarative Test Driven Development approach through Invariant's
        * `blackbox_pyteal_example4()`: Using PyTealDryRunExecutor to debug an ABIReturnSubroutine with an app, logic sig and csv reports
        """

        return self._pyteal_lambda()

    def _arg_prep_n_call(self, i, p):
        subdef = self.subr.subroutine.subroutine
        arg_names = subdef.arguments()
        name = arg_names[i]
        arg_expr = Txn.application_args[i] if self.mode == Mode.Application else Arg(i)
        if p == TealType.uint64:
            arg_expr = Btoi(arg_expr)

        if name in subdef.by_ref_args:
            arg_var = ScratchVar(p)
            prep = arg_var.store(arg_expr)
        elif name in subdef.abi_args:
            arg_var = p.new_instance()
            prep = arg_var.decode(arg_expr)
        else:
            arg_var = arg_expr
            prep = None
        return prep, arg_var

    def _prepare_n_calls(self):
        preps_n_calls = [
            *(self._arg_prep_n_call(i, p) for i, p in enumerate(self.input_types))
        ]
        preps, calls = zip(*preps_n_calls) if preps_n_calls else ([], [])
        preps = [p for p in preps if p]
        return preps, calls

    def _handle_SubroutineFnWrapper(self):
        subdef = self.subr.subroutine.subroutine

        def subr_caller():
            preps, calls = self._prepare_n_calls()
            invocation = self.subr(*calls)
            if preps:
                return Seq(*(preps + [invocation]))
            return invocation

        def make_return(e):
            if e.type_of() == TealType.uint64:
                return e
            if e.type_of() == TealType.bytes:
                return Len(e)
            if e.type_of() == TealType.anytype:
                x = ScratchVar(TealType.anytype)
                return Seq(x.store(e), Int(1337))
            # TealType.none:
            return Seq(e, Int(1337))

        def make_log(e):
            if e.type_of() == TealType.uint64:
                return Log(Itob(e))
            if e.type_of() == TealType.bytes:
                return Log(e)
            return Log(Bytes("nada"))

        if self.mode == Mode.Signature:

            def approval():
                return make_return(subr_caller())

        else:

            def approval():
                if subdef.return_type == TealType.none:
                    result = ScratchVar(TealType.uint64)
                    part1 = [subr_caller(), result.store(Int(1337))]
                else:
                    result = ScratchVar(subdef.return_type)
                    part1 = [result.store(subr_caller())]

                part2 = [make_log(result.load()), make_return(result.load())]
                return Seq(*(part1 + part2))

        return approval

    def _handle_ABIReturnSubroutine(self):
        output = None
        if self.subr.subroutine.output_kwarg_info:
            output = self.subr.subroutine.output_kwarg_info.abi_type.new_instance()

        def approval():
            preps, calls = self._prepare_n_calls()

            # when @ABIReturnSubroutine is void:
            #   invocation is an Expr of TealType.none
            # otherwise:
            #   it is a ComputedValue
            invocation = self.subr(*calls)
            if output:
                invocation = output.set(invocation)
                if self.mode == Mode.Signature:
                    results = [invocation, Pop(output.encode()), Int(1)]
                else:
                    results = [invocation, abi.MethodReturn(output), Int(1)]
            else:
                results = [invocation, Int(1)]

            return Seq(*(preps + results))

        return approval

    def compile(self, version: int, assemble_constants: bool = False) -> str:
        return _MatchMode(
            app_case=lambda: compileTeal(
                self.program(),
                self.mode,
                version=version,
                assembleConstants=assemble_constants,
            ),
            signature_case=lambda: compileTeal(
                self.program(),
                self.mode,
                version=version,
                assembleConstants=assemble_constants,
            ),
        )(self.mode)

    def dryrun_on_sequence(
        self,
        inputs: list[Sequence[str | int]],
        compiler_version=6,
    ) -> list[DryRunInspector]:
        return _MatchMode(
            app_case=lambda: DryRunExecutor.dryrun_app_on_sequence(
                algod_with_assertion(),
                self.compile(compiler_version),
                inputs,
                self.abi_argument_types(),
                self.abi_return_type(),
            ),
            signature_case=lambda: DryRunExecutor.dryrun_logicsig_on_sequence(
                algod_with_assertion(),
                self.compile(compiler_version),
                inputs,
                self.abi_argument_types(),
                self.abi_return_type(),
            ),
        )(self.mode)

    def dryrun(
        self,
        args: Sequence[bytes | str | int],
        compiler_version=6,
    ) -> DryRunInspector:
        return _MatchMode(
            app_case=lambda: DryRunExecutor.dryrun_app(
                algod_with_assertion(),
                self.compile(compiler_version),
                args,
                self.abi_argument_types(),
                self.abi_return_type(),
            ),
            signature_case=lambda: DryRunExecutor.dryrun_logicsig(
                algod_with_assertion(),
                self.compile(compiler_version),
                args,
                self.abi_argument_types(),
                self.abi_return_type(),
            ),
        )(self.mode)



import pytest

import pyteal as pt

avm6Options = pt.CompileOptions(version=6)


def test_gtxn_invalid():
    for f, e in [
        (lambda: pt.Gtxn[-1], pt.TealInputError),
        (lambda: pt.Gtxn[pt.MAX_GROUP_SIZE + 1], pt.TealInputError),
        (lambda: pt.Gtxn[pt.Pop(pt.Int(0))], pt.TealTypeError),
        (lambda: pt.Gtxn[pt.Bytes("index")], pt.TealTypeError),
    ]:
        with pytest.raises(e):
            f()


def test_gtxn_expr_invalid():
    for f, e in [
        (
            lambda: pt.GtxnExpr(pt.Assert(pt.Int(1)), pt.TxnField.sender),
            pt.TealTypeError,
        ),
    ]:
        with pytest.raises(e):
            f()


def test_gtxn_expr_valid():
    [
        e.__teal__(avm6Options)
        for e in [
            pt.GtxnExpr(1, pt.TxnField.sender),
            pt.GtxnExpr(pt.Int(1), pt.TxnField.sender),
        ]
    ]


def test_gtxna_expr_invalid():
    for f, e in [
        (
            lambda: pt.GtxnaExpr("Invalid_type", pt.TxnField.assets, 1),
            pt.TealInputError,
        ),
        (
            lambda: pt.GtxnaExpr(1, pt.TxnField.assets, "Invalid_type"),
            pt.TealInputError,
        ),
        (
            lambda: pt.GtxnaExpr(pt.Assert(pt.Int(1)), pt.TxnField.assets, 1),
            pt.TealTypeError,
        ),
        (
            lambda: pt.GtxnaExpr(1, pt.TxnField.assets, pt.Assert(pt.Int(1))),
            pt.TealTypeError,
        ),
    ]:
        with pytest.raises(e):
            f()


def test_gtxna_expr_valid():
    [
        e.__teal__(avm6Options)
        for e in [
            pt.GtxnaExpr(1, pt.TxnField.assets, 1),
            pt.GtxnaExpr(pt.Int(1), pt.TxnField.assets, pt.Int(1)),
        ]
    ]


# txn_test.py performs additional testing



from typing import TYPE_CHECKING
from enum import Enum

from pyteal.types import TealType
from pyteal.errors import verifyFieldVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.leafexpr import LeafExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class GlobalField(Enum):
    min_txn_fee = (0, "MinTxnFee", TealType.uint64, 2)
    min_balance = (1, "MinBalance", TealType.uint64, 2)
    max_txn_life = (2, "MaxTxnLife", TealType.uint64, 2)
    zero_address = (3, "ZeroAddress", TealType.bytes, 2)
    group_size = (4, "GroupSize", TealType.uint64, 2)
    logic_sig_version = (5, "LogicSigVersion", TealType.uint64, 2)
    round = (6, "Round", TealType.uint64, 2)
    latest_timestamp = (7, "LatestTimestamp", TealType.uint64, 2)
    current_app_id = (8, "CurrentApplicationID", TealType.uint64, 2)
    creator_address = (9, "CreatorAddress", TealType.bytes, 3)
    current_app_address = (10, "CurrentApplicationAddress", TealType.bytes, 5)
    group_id = (11, "GroupID", TealType.bytes, 5)
    opcode_budget = (12, "OpcodeBudget", TealType.uint64, 6)
    caller_app_id = (13, "CallerApplicationID", TealType.uint64, 6)
    caller_app_address = (14, "CallerApplicationAddress", TealType.bytes, 6)

    def __init__(self, id: int, name: str, type: TealType, min_version: int) -> None:
        self.id = id
        self.arg_name = name
        self.ret_type = type
        self.min_version = min_version

    def type_of(self) -> TealType:
        return self.ret_type


GlobalField.__module__ = "pyteal"


class Global(LeafExpr):
    """An expression that accesses a global property."""

    def __init__(self, field: GlobalField) -> None:
        super().__init__()
        self.field = field

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(self.field.arg_name, self.field.min_version, options.version)

        op = TealOp(self, Op.global_, self.field.arg_name)
        return TealBlock.FromOp(options, op)

    def __str__(self):
        return "(Global {})".format(self.field.arg_name)

    def type_of(self):
        return self.field.type_of()

    @classmethod
    def min_txn_fee(cls) -> "Global":
        """Get the minimum transaction fee in micro Algos."""
        return cls(GlobalField.min_txn_fee)

    @classmethod
    def min_balance(cls) -> "Global":
        """Get the minimum balance in micro Algos."""
        return cls(GlobalField.min_balance)

    @classmethod
    def max_txn_life(cls) -> "Global":
        """Get the maximum number of rounds a transaction can have."""
        return cls(GlobalField.max_txn_life)

    @classmethod
    def zero_address(cls) -> "Global":
        """Get the 32 byte zero address."""
        return cls(GlobalField.zero_address)

    @classmethod
    def group_size(cls) -> "Global":
        """Get the number of transactions in this atomic transaction group.

        This will be at least 1.
        """
        return cls(GlobalField.group_size)

    @classmethod
    def logic_sig_version(cls) -> "Global":
        """Get the maximum supported program version."""
        return cls(GlobalField.logic_sig_version)

    @classmethod
    def round(cls) -> "Global":
        """Get the current round number."""
        return cls(GlobalField.round)

    @classmethod
    def latest_timestamp(cls) -> "Global":
        """Get the latest confirmed block UNIX timestamp.

        Fails if negative."""
        return cls(GlobalField.latest_timestamp)

    @classmethod
    def current_application_id(cls) -> "Global":
        """Get the ID of the current application executing.

        Fails during Signature mode."""
        return cls(GlobalField.current_app_id)

    @classmethod
    def creator_address(cls) -> "Global":
        """Address of the creator of the current application.

        Fails during Signature mode. Requires program version 3 or higher."""
        return cls(GlobalField.creator_address)

    @classmethod
    def current_application_address(cls) -> "Global":
        """Get the address of that the current application controls.

        Fails during Signature mode. Requires program version 5 or higher."""
        return cls(GlobalField.current_app_address)

    @classmethod
    def group_id(cls) -> "Global":
        """Get the ID of the current transaction group.

        If the current transaction is not part of a group, this will return 32 zero bytes.

        Requires program version 5 or higher."""
        return cls(GlobalField.group_id)

    @classmethod
    def opcode_budget(cls) -> "Global":
        """Get the remaining opcode execution budget

        Requires program version 6 or higher."""
        return cls(GlobalField.opcode_budget)

    @classmethod
    def caller_app_id(cls) -> "Global":
        """Get the id of the app that submitted the InnerTransaction that triggered this app to execute.

        If not called from another app, this will return 0

        Requires program version 6 or higher."""
        return cls(GlobalField.caller_app_id)

    @classmethod
    def caller_app_address(cls) -> "Global":
        """Get the address of the app that submitted the InnerTransaction that triggered this app to execute.

        If not called from another app, this will return the ZeroAddress

        Requires program version 6 or higher."""
        return cls(GlobalField.caller_app_address)


Global.__module__ = "pyteal"



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *

"""Basic Bank"""


def bank_for_account(receiver):
    """Only allow receiver to withdraw funds from this contract account.

    Args:
        receiver (str): Base 32 Algorand address of the receiver.
    """

    is_payment = Txn.type_enum() == TxnType.Payment
    is_single_tx = Global.group_size() == Int(1)
    is_correct_receiver = Txn.receiver() == Addr(receiver)
    no_close_out_addr = Txn.close_remainder_to() == Global.zero_address()
    no_rekey_addr = Txn.rekey_to() == Global.zero_address()
    acceptable_fee = Txn.fee() <= Int(1000)

    return And(
        is_payment,
        is_single_tx,
        is_correct_receiver,
        no_close_out_addr,
        no_rekey_addr,
        acceptable_fee,
    )


if __name__ == "__main__":
    program = bank_for_account(
        "ZZAF5ARA4MEC5PVDOP64JM5O5MQST63Q2KOY2FLYFLXXD3PFSNJJBYAFZM"
    )
    print(compileTeal(program, mode=Mode.Signature, version=3))



from typing import TYPE_CHECKING

from pyteal.types import TealType
from pyteal.errors import TealCompileError
from pyteal.ast.expr import Expr
from pyteal.ir import TealSimpleBlock


if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Continue(Expr):
    """A continue expression"""

    def __init__(self) -> None:
        """Create a new continue expression.

        This operation is only permitted in a loop.

        """
        super().__init__()

    def __str__(self) -> str:
        return "continue"

    def __teal__(self, options: "CompileOptions"):
        if not options.isInLoop():
            raise TealCompileError("continue is only allowed in a loop", self)

        start = TealSimpleBlock([])
        options.addLoopContinueBlock(start)

        return start, start

    def type_of(self):
        return TealType.none

    def has_return(self):
        return False


Continue.__module__ = "pyteal"



from typing import Final, Generic, Literal, Sequence, TypeVar, Union, cast

from pyteal.errors import TealInputError
from pyteal.ast.assert_ import Assert
from pyteal.ast.expr import Expr
from pyteal.ast.int import Int
from pyteal.ast.bytes import Bytes
from pyteal.ast.seq import Seq
from pyteal.ast.unaryexpr import Len

from pyteal.ast.abi.type import ComputedValue, TypeSpec, BaseType
from pyteal.ast.abi.bool import BoolTypeSpec, _bool_sequence_length
from pyteal.ast.abi.uint import Byte, ByteTypeSpec
from pyteal.ast.abi.array_base import ArrayTypeSpec, Array, ArrayElement

T = TypeVar("T", bound=BaseType)
N = TypeVar("N", bound=int)


class StaticArrayTypeSpec(ArrayTypeSpec[T], Generic[T, N]):
    def __init__(self, value_type_spec: TypeSpec, array_length: int) -> None:
        super().__init__(value_type_spec)
        if not isinstance(array_length, int) or array_length < 0:
            raise TypeError(f"Unsupported StaticArray length: {array_length}")

        # Casts to `int` to handle downstream usage where value is a subclass of int like `IntEnum`.
        self.array_length: Final = int(array_length)

    def new_instance(self) -> "StaticArray[T, N]":
        return StaticArray(self)

    def annotation_type(self) -> type["StaticArray[T, N]"]:
        return StaticArray[  # type: ignore[misc]
            self.value_spec.annotation_type(), Literal[self.array_length]  # type: ignore
        ]

    def length_static(self) -> int:
        """Get the size of this static array type.

        Returns:
            A Python integer that represents the static array length.
        """
        return self.array_length

    def is_length_dynamic(self) -> bool:
        return False

    def is_dynamic(self) -> bool:
        return self.value_type_spec().is_dynamic()

    def byte_length_static(self) -> int:
        if self.is_dynamic():
            raise ValueError("Type is dynamic")

        value_type = self.value_type_spec()
        length = self.length_static()

        if value_type == BoolTypeSpec():
            return _bool_sequence_length(length)
        return length * value_type.byte_length_static()

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, StaticArrayTypeSpec)
            and self.value_type_spec() == other.value_type_spec()
            and self.length_static() == other.length_static()
        )

    def __str__(self) -> str:
        return f"{self.value_type_spec()}[{self.length_static()}]"


StaticArrayTypeSpec.__module__ = "pyteal.abi"


class StaticArray(Array[T], Generic[T, N]):
    """The class that represents ABI static array type."""

    def __init__(self, array_type_spec: StaticArrayTypeSpec[T, N]) -> None:
        super().__init__(array_type_spec)

    def type_spec(self) -> StaticArrayTypeSpec[T, N]:
        return cast(StaticArrayTypeSpec[T, N], super().type_spec())

    def set(
        self,
        values: Union[
            Sequence[T], "StaticArray[T, N]", ComputedValue["StaticArray[T, N]"]
        ],
    ) -> Expr:
        """Set the elements of this StaticArray to the input values.

        The behavior of this method depends on the input argument type:

            * :code:`Sequence[T]`: set the elements of this StaticArray to those contained in this Python sequence (e.g. a list or tuple). A compiler error will occur if any element in the sequence does not match this StaticArray's element type, or if the sequence length does not equal this StaticArray's length.
            * :code:`StaticArray[T, N]`: copy the elements from another StaticArray. The argument's element type and length must exactly match this StaticArray's element type and length, otherwise an error will occur.
            * :code:`ComputedValue[StaticArray[T, N]]`: copy the elements from a StaticArray produced by a ComputedValue. The element type and length produced by the ComputedValue must exactly match this StaticArray's element type and length, otherwise an error will occur.

        Args:
            values: The new elements this StaticArray should have. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this StaticArray.
        """
        if isinstance(values, ComputedValue):
            return self._set_with_computed_type(values)
        elif isinstance(values, BaseType):
            if self.type_spec() != values.type_spec():
                raise TealInputError(
                    f"Cannot assign type {values.type_spec()} to {self.type_spec()}"
                )
            return self.stored_value.store(values.encode())

        if self.type_spec().length_static() != len(values):
            raise TealInputError(
                f"Incorrect length for values. Expected {self.type_spec()}, got {len(values)}"
            )
        return super().set(values)

    def length(self) -> Expr:
        """Get the element number of this ABI static array.

        Returns:
            A PyTeal expression that represents the static array length.
        """
        return Int(self.type_spec().length_static())

    def __getitem__(self, index: Union[int, Expr]) -> "ArrayElement[T]":
        """Retrieve an element by its index in this StaticArray.

        Indexes start at 0.

        Args:
            index: either a Python integer or a PyTeal expression that evaluates to a TealType.uint64.
                If a Python integer is used, this function will raise an error if its value is negative
                or if the index is equal to or greater than the length of this StaticArray. If a PyTeal
                expression is used, the program will fail at runtime if the index is outside of the
                bounds of this StaticArray.

        Returns:
            An ArrayElement that corresponds to the element at the given index. This type is a ComputedValue.
        """
        if type(index) is int and index >= self.type_spec().length_static():
            raise TealInputError(f"Index out of bounds: {index}")
        return super().__getitem__(index)


StaticArray.__module__ = "pyteal.abi"


class StaticBytesTypeSpec(StaticArrayTypeSpec[Byte, N], Generic[N]):
    def __init__(self, array_length: int) -> None:
        super().__init__(ByteTypeSpec(), array_length)

    def new_instance(self) -> "StaticBytes[N]":
        return StaticBytes(self)

    def annotation_type(self) -> type["StaticBytes[N]"]:
        return StaticBytes[  # type: ignore[misc]
            Literal[self.array_length]  # type: ignore
        ]


StaticBytesTypeSpec.__module__ = "pyteal.abi"


class StaticBytes(StaticArray[Byte, N], Generic[N]):
    """The convenience class that represents ABI static byte array."""

    def __init__(self, array_type_spec: StaticBytesTypeSpec[N]) -> None:
        super().__init__(array_type_spec)

    def type_spec(self) -> StaticBytesTypeSpec:
        return cast(StaticBytesTypeSpec[N], super().type_spec())

    def set(
        self,
        values: Union[
            bytes,
            bytearray,
            Expr,
            Sequence[Byte],
            StaticArray[Byte, N],
            ComputedValue[StaticArray[Byte, N]],
        ],
    ) -> Expr:
        """Set the elements of this StaticBytes to the input values.

        The behavior of this method depends on the input argument type:

            * :code:`bytes`: set the value to the Python byte string.
            * :code:`bytearray`: set the value to the Python byte array.
            * :code:`Expr`: set the value to the result of a PyTeal expression, which must evaluate to a TealType.bytes.
            * :code:`Sequence[Byte]`: set the bytes of this String to those contained in this Python sequence (e.g. a list or tuple).
            * :code:`StaticArray[Byte, N]`: copy the bytes from another StaticArray. The argument's element type and length must exactly match Byte and this StaticBytes' length, otherwise an error will occur.
            * :code:`ComputedValue[StaticArray[Byte, N]]`: copy the bytes from a StaticArray produced by a ComputedType. The argument's element type and length must exactly match Byte and this StaticBytes' length, otherwise an error will occur.

        Args:
            values: The new elements this StaticBytes should have. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this StaticBytes.
        """
        match values:
            case bytes() | bytearray():
                if len(values) != self.type_spec().length_static():
                    raise TealInputError(
                        f"Got bytes with length {len(values)}, expect {self.type_spec().length_static()}"
                    )
                return self.stored_value.store(Bytes(values))
            case Expr():
                return Seq(
                    self.stored_value.store(values),
                    Assert(self.length() == Len(self.stored_value.load())),
                )

        return super().set(values)

    def get(self) -> Expr:
        """Get the byte encoding of this StaticBytes.

        Returns:
            A Pyteal expression that loads byte encoding of this StaticBytes.
        """
        return self.stored_value.load()


StaticBytes.__module__ = "pyteal.abi"



from typing import (
    Any,
    Literal,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

import algosdk.abi

from pyteal.errors import TealInputError
from pyteal.ast.expr import Expr
from pyteal.ast.int import Int
from pyteal.ast.substring import Extract, Substring, Suffix
from pyteal.ast.abi.type import TypeSpec, BaseType


def substring_for_decoding(
    encoded: Expr,
    *,
    start_index: Expr = None,
    end_index: Expr = None,
    length: Expr = None,
) -> Expr:
    """A helper function for getting the substring to decode according to the rules of BaseType.decode."""
    if length is not None and end_index is not None:
        raise TealInputError("length and end_index are mutually exclusive arguments")

    if start_index is not None:
        if length is not None:
            # substring from start_index to start_index + length
            return Extract(encoded, start_index, length)

        if end_index is not None:
            # substring from start_index to end_index
            return Substring(encoded, start_index, end_index)

        # substring from start_index to end of string
        return Suffix(encoded, start_index)

    if length is not None:
        # substring from 0 to length
        return Extract(encoded, Int(0), length)

    if end_index is not None:
        # substring from 0 to end_index
        return Substring(encoded, Int(0), end_index)

    # the entire string
    return encoded


def int_literal_from_annotation(annotation: Any) -> int:
    """Extract an integer from a Literal type annotation.

    Args:
        annotation: A Literal type annotation. E.g., `Literal[4]`. This must contain only a single
            integer value.

    Returns:
        The integer that the Literal represents.
    """
    origin = get_origin(annotation)
    args = get_args(annotation)

    if origin is not Literal:
        raise TypeError("Expected literal for argument. Got: {}".format(origin))

    if len(args) != 1 or type(args[0]) is not int:
        raise TypeError(
            "Expected single integer argument for Literal. Got: {}".format(args)
        )

    return args[0]


def type_spec_from_annotation(annotation: Any) -> TypeSpec:
    """Convert an ABI type annotation into the corresponding TypeSpec.

    For example, calling this function with the input `abi.StaticArray[abi.Bool, Literal[5]]` would
    return `abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 5)`.

    Args:
        annotation: An annotation representing an ABI type instance.

    Raises:
        TypeError: if the input annotation does not represent a valid ABI type instance or its
            arguments are invalid.

    Returns:
        The TypeSpec that corresponds to the input annotation.
    """
    from pyteal.ast.abi.bool import BoolTypeSpec, Bool
    from pyteal.ast.abi.uint import (
        ByteTypeSpec,
        Byte,
        Uint8TypeSpec,
        Uint8,
        Uint16TypeSpec,
        Uint16,
        Uint32TypeSpec,
        Uint32,
        Uint64TypeSpec,
        Uint64,
    )
    from pyteal.ast.abi.array_dynamic import (
        DynamicArrayTypeSpec,
        DynamicArray,
        DynamicBytesTypeSpec,
        DynamicBytes,
    )
    from pyteal.ast.abi.array_static import (
        StaticArrayTypeSpec,
        StaticArray,
        StaticBytesTypeSpec,
        StaticBytes,
    )
    from pyteal.ast.abi.tuple import (
        TupleTypeSpec,
        Tuple,
        Tuple0,
        Tuple1,
        Tuple2,
        Tuple3,
        Tuple4,
        Tuple5,
        NamedTuple,
        NamedTupleTypeSpec,
    )
    from pyteal.ast.abi.string import StringTypeSpec, String
    from pyteal.ast.abi.address import AddressTypeSpec, Address
    from pyteal.ast.abi.transaction import (
        Transaction,
        TransactionTypeSpec,
        PaymentTransaction,
        PaymentTransactionTypeSpec,
        KeyRegisterTransaction,
        KeyRegisterTransactionTypeSpec,
        AssetConfigTransaction,
        AssetConfigTransactionTypeSpec,
        AssetFreezeTransaction,
        AssetFreezeTransactionTypeSpec,
        AssetTransferTransaction,
        AssetTransferTransactionTypeSpec,
        ApplicationCallTransaction,
        ApplicationCallTransactionTypeSpec,
    )
    from pyteal.ast.abi.reference_type import (
        AccountTypeSpec,
        Account,
        AssetTypeSpec,
        Asset,
        ApplicationTypeSpec,
        Application,
    )

    origin = get_origin(annotation)
    if origin is None:
        origin = annotation

    args = get_args(annotation)

    if origin is Account:
        if len(args) != 0:
            raise TypeError("Account expects 0 arguments. Got: {}".format(args))
        return AccountTypeSpec()

    if origin is Asset:
        if len(args) != 0:
            raise TypeError("Asset expects 0 arguments. Got: {}".format(args))
        return AssetTypeSpec()

    if origin is Application:
        if len(args) != 0:
            raise TypeError("Application expects 0 arguments. Got: {}".format(args))
        return ApplicationTypeSpec()

    if origin is Bool:
        if len(args) != 0:
            raise TypeError("Bool expects 0 type arguments. Got: {}".format(args))
        return BoolTypeSpec()

    if origin is Byte:
        if len(args) != 0:
            raise TypeError("Byte expects 0 type arguments. Got: {}".format(args))
        return ByteTypeSpec()

    if origin is Uint8:
        if len(args) != 0:
            raise TypeError("Uint8 expects 0 type arguments. Got: {}".format(args))
        return Uint8TypeSpec()

    if origin is Uint16:
        if len(args) != 0:
            raise TypeError("Uint16 expects 0 type arguments. Got: {}".format(args))
        return Uint16TypeSpec()

    if origin is Uint32:
        if len(args) != 0:
            raise TypeError("Uint32 expects 0 type arguments. Got: {}".format(args))
        return Uint32TypeSpec()

    if origin is Uint64:
        if len(args) != 0:
            raise TypeError("Uint64 expects 0 type arguments. Got: {}".format(args))
        return Uint64TypeSpec()

    if origin is String:
        if len(args) != 0:
            raise TypeError("String expects 0 arguments. Got: {}".format(args))
        return StringTypeSpec()

    if origin is Address:
        if len(args) != 0:
            raise TypeError("Address expects 0 arguments. Got: {}".format(args))
        return AddressTypeSpec()

    if origin is DynamicBytes:
        if len(args) != 0:
            raise TypeError(f"DynamicBytes expect 0 type argument. Got: {args}")
        return DynamicBytesTypeSpec()

    if origin is DynamicArray:
        if len(args) != 1:
            raise TypeError(
                "DynamicArray expects 1 type argument. Got: {}".format(args)
            )
        value_type_spec = type_spec_from_annotation(args[0])
        return DynamicArrayTypeSpec(value_type_spec)

    if origin is StaticBytes:
        if len(args) != 1:
            raise TypeError(f"StaticBytes expect 1 type argument. Got: {args}")
        array_length = int_literal_from_annotation(args[0])
        return StaticBytesTypeSpec(array_length)

    if origin is StaticArray:
        if len(args) != 2:
            raise TypeError("StaticArray expects 1 type argument. Got: {}".format(args))
        value_type_spec = type_spec_from_annotation(args[0])
        array_length = int_literal_from_annotation(args[1])
        return StaticArrayTypeSpec(value_type_spec, array_length)

    if origin is Tuple:
        return TupleTypeSpec(*(type_spec_from_annotation(arg) for arg in args))

    if issubclass(origin, NamedTuple):
        return cast(NamedTupleTypeSpec, origin().type_spec())

    if origin is Tuple0:
        if len(args) != 0:
            raise TypeError("Tuple0 expects 0 type arguments. Got: {}".format(args))
        return TupleTypeSpec()

    if origin is Tuple1:
        if len(args) != 1:
            raise TypeError("Tuple1 expects 1 type argument. Got: {}".format(args))
        return TupleTypeSpec(*(type_spec_from_annotation(arg) for arg in args))

    if origin is Tuple2:
        if len(args) != 2:
            raise TypeError("Tuple2 expects 2 type arguments. Got: {}".format(args))
        return TupleTypeSpec(*(type_spec_from_annotation(arg) for arg in args))

    if origin is Tuple3:
        if len(args) != 3:
            raise TypeError("Tuple3 expects 3 type arguments. Got: {}".format(args))
        return TupleTypeSpec(*(type_spec_from_annotation(arg) for arg in args))

    if origin is Tuple4:
        if len(args) != 4:
            raise TypeError("Tuple4 expects 4 type arguments. Got: {}".format(args))
        return TupleTypeSpec(*(type_spec_from_annotation(arg) for arg in args))

    if origin is Tuple5:
        if len(args) != 5:
            raise TypeError("Tuple5 expects 5 type arguments. Got: {}".format(args))
        return TupleTypeSpec(*(type_spec_from_annotation(arg) for arg in args))

    if origin is Transaction:
        if len(args) != 0:
            raise TypeError("Transaction expects 0 type arguments. Got {}".format(args))
        return TransactionTypeSpec()

    if origin is PaymentTransaction:
        if len(args) != 0:
            raise TypeError(
                "PaymentTransaction expects 0 type arguments. Got {}".format(args)
            )
        return PaymentTransactionTypeSpec()

    if origin is KeyRegisterTransaction:
        if len(args) != 0:
            raise TypeError(
                "KeyRegisterTransaction expects 0 type arguments. Got {}".format(args)
            )
        return KeyRegisterTransactionTypeSpec()

    if origin is AssetConfigTransaction:
        if len(args) != 0:
            raise TypeError(
                "AssetConfigTransaction expects 0 type arguments. Got {}".format(args)
            )
        return AssetConfigTransactionTypeSpec()

    if origin is AssetFreezeTransaction:
        if len(args) != 0:
            raise TypeError(
                "AssetFreezeTransaction expects 0 type arguments. Got {}".format(args)
            )
        return AssetFreezeTransactionTypeSpec()

    if origin is AssetTransferTransaction:
        if len(args) != 0:
            raise TypeError(
                "AssetTransferTransaction expects 0 type arguments. Got {}".format(args)
            )
        return AssetTransferTransactionTypeSpec()

    if origin is ApplicationCallTransaction:
        if len(args) != 0:
            raise TypeError(
                "ApplicationCallTransaction expects 0 type arguments. Got {}".format(
                    args
                )
            )
        return ApplicationCallTransactionTypeSpec()

    raise TypeError("Unknown annotation origin: {}".format(origin))


T = TypeVar("T", bound=BaseType)


def contains_type_spec(ts: TypeSpec, targets: Sequence[TypeSpec]) -> bool:
    from pyteal.ast.abi.array_dynamic import DynamicArrayTypeSpec
    from pyteal.ast.abi.array_static import StaticArrayTypeSpec
    from pyteal.ast.abi.tuple import TupleTypeSpec

    stack: list[TypeSpec] = [ts]

    while stack:
        current = stack.pop()
        if current in targets:
            return True

        match current:
            case TupleTypeSpec():
                stack.extend(current.value_type_specs())
            case DynamicArrayTypeSpec():
                stack.append(current.value_type_spec())
            case StaticArrayTypeSpec():
                stack.append(current.value_type_spec())

    return False


def size_of(t: type[T]) -> int:
    """Get the size in bytes of an ABI type. Must be a static type"""

    ts = type_spec_from_annotation(t)
    if ts.is_dynamic():
        raise TealInputError("Cannot get size of dynamic type")

    return ts.byte_length_static()


def make(t: type[T]) -> T:
    """Create a new instance of an ABI type. The type to create is determined by the input argument,
    which must be a fully-specified type's class. Fully-specified means that every generic argument
    is given a value.

    For example:
        .. code-block:: python

                # both of these are equivalent
                a = abi.make(abi.Tuple2[abi.Uint64, abi.StaticArray[abi.Bool, Literal[8]]])
                b = abi.TupleTypeSpec(abi.Uint64TypeSpec(), abi.StaticArrayTypeSpec(abi.BoolTypeSpec(), 8))

    This is purely a convenience method over instantiating the type directly, which can be cumbersome
    due to the lengthy TypeSpec class names.

    Args:
        t: A fully-specified subclass of abi.BaseType.

    Returns:
        A new instance of the given type class.
    """
    return cast(T, type_spec_from_annotation(t).new_instance())


def algosdk_from_type_spec(t: TypeSpec) -> algosdk.abi.ABIType:
    from pyteal.ast.abi import ReferenceTypeSpecs, TransactionTypeSpecs

    if t in TransactionTypeSpecs:
        raise TealInputError(
            f"cannot map ABI transaction type spec {t!r} to an appropriate algosdk ABI type"
        )

    if t in ReferenceTypeSpecs:
        raise TealInputError(
            f"cannot map ABI reference type spec {t!r} to an appropriate algosdk ABI type"
        )

    return algosdk.abi.ABIType.from_string(str(t))


def algosdk_from_annotation(t: type[T]) -> algosdk.abi.ABIType:
    return algosdk_from_type_spec(type_spec_from_annotation(t))


def type_spec_from_algosdk(t: Union[algosdk.abi.ABIType, str]) -> TypeSpec:

    from pyteal.ast.abi.reference_type import ReferenceTypeSpecs
    from pyteal.ast.abi.transaction import TransactionTypeSpecs

    from pyteal.ast.abi.array_dynamic import DynamicArrayTypeSpec
    from pyteal.ast.abi.array_static import StaticArrayTypeSpec
    from pyteal.ast.abi.tuple import TupleTypeSpec
    from pyteal.ast.abi.bool import BoolTypeSpec
    from pyteal.ast.abi.string import StringTypeSpec
    from pyteal.ast.abi.address import AddressTypeSpec
    from pyteal.ast.abi.uint import (
        ByteTypeSpec,
        Uint8TypeSpec,
        Uint16TypeSpec,
        Uint32TypeSpec,
        Uint64TypeSpec,
    )

    match t:
        # Currently reference and transaction types are only strings
        case str():
            if algosdk.abi.is_abi_reference_type(t):
                ref_dict: dict[str, TypeSpec] = {
                    str(rts): rts for rts in ReferenceTypeSpecs
                }
                if t in ref_dict:
                    return ref_dict[t]
                else:
                    raise TealInputError(f"Invalid reference type: {t}")

            elif algosdk.abi.is_abi_transaction_type(t):
                txn_dict: dict[str, TypeSpec] = {
                    str(tts): tts for tts in TransactionTypeSpecs
                }
                if t in txn_dict:
                    return txn_dict[t]
                else:
                    raise TealInputError(f"Invalid transaction type: {t}")
            else:
                raise TealInputError(f"Invalid ABI type: {t}")

        case algosdk.abi.ABIType():
            match t:
                case algosdk.abi.ArrayDynamicType():
                    return DynamicArrayTypeSpec(type_spec_from_algosdk(t.child_type))
                case algosdk.abi.ArrayStaticType():
                    return StaticArrayTypeSpec(
                        type_spec_from_algosdk(t.child_type), t.static_length
                    )
                case algosdk.abi.TupleType():
                    return TupleTypeSpec(
                        *[type_spec_from_algosdk(ct) for ct in t.child_types]
                    )
                case algosdk.abi.UintType():
                    match t.bit_size:
                        case 8:
                            return Uint8TypeSpec()
                        case 16:
                            return Uint16TypeSpec()
                        case 32:
                            return Uint32TypeSpec()
                        case 64:
                            return Uint64TypeSpec()
                case algosdk.abi.ByteType():
                    return ByteTypeSpec()
                case algosdk.abi.BoolType():
                    return BoolTypeSpec()
                case algosdk.abi.StringType():
                    return StringTypeSpec()
                case algosdk.abi.AddressType():
                    return AddressTypeSpec()
                case algosdk.abi.UfixedType():
                    raise TealInputError("Ufixed not supported")

    raise TealInputError(f"Invalid Type: {t}")


def type_specs_from_signature(sig: str) -> tuple[list[TypeSpec], Optional[TypeSpec]]:
    sdk_method = algosdk.abi.Method.from_signature(sig)

    return_type = None
    if sdk_method.returns.type != algosdk.abi.Returns.VOID:
        return_type = type_spec_from_algosdk(sdk_method.returns.type)

    return [type_spec_from_algosdk(arg.type) for arg in sdk_method.args], return_type


def type_spec_is_assignable_to(a: TypeSpec, b: TypeSpec) -> bool:
    """Decides if the value of type :code:`a` can be assigned to or interpreted as another value of type :code:`b`.

    This method return true if and only if all of the following properties hold:

        * value of type :code:`a` has identical encoding as value of type :code:`b`
        * type :code:`b` is as general as, or more general than type :code:`a`

    For `abi.NamedTuple`, we allow mutual assigning between `abi.Tuple` and `abi.NamedTuple`.
    But between `abi.NamedTuple`, we only return true when the type specs are identical, or we cannot compare against generality.

    Some examples are illustrated as following:

    =========================== =========================== ============= ====================================================================
    Type :code:`a`              Type :code:`b`              Assignable?   Reason
    =========================== =========================== ============= ====================================================================
    :code:`DynamicArray[Byte]`  :code:`DynamicBytes`        :code:`True`  :code:`DynamicBytes` is as general as :code:`DynamicArray[Byte]`
    :code:`DynamicBytes`        :code:`DynamicArray[Byte]`  :code:`True`  :code:`DynamicArray[Byte]` is as general as :code:`DynamicBytes`
    :code:`StaticArray[Byte,N]` :code:`StaticBytes[N]`      :code:`True`  :code:`StaticBytes[N]` is as general as :code:`StaticArray[Byte,N]`
    :code:`StaticBytes[N]`      :code:`StaticArray[Byte,N]` :code:`True`  :code:`StaticArray[Byte,N]` is as general as :code:`StaticBytes[N]`
    :code:`String`              :code:`DynamicBytes`        :code:`True`  :code:`DynamicBytes` is more general than :code:`String`
    :code:`DynamicBytes`        :code:`String`              :code:`False` :code:`String` is more specific than :code:`DynamicBytes`
    :code:`Address`             :code:`StaticBytes[32]`     :code:`True`  :code:`StaticBytes[32]` is more general than :code:`Address`
    :code:`StaticBytes[32]`     :code:`Address`             :code:`False` :code:`Address` is more specific than :code:`StaticBytes[32]`
    :code:`PaymentTransaction`  :code:`Transaction`         :code:`True`  :code:`Transaction` is more general than :code:`PaymentTransaction`
    :code:`Transaction`         :code:`PaymentTransaction`  :code:`False` :code:`PaymentTransaction` is more specific than :code`Transaction`
    :code:`Uint8`               :code:`Byte`                :code:`True`  :code:`Uint8` is as general as :code:`Byte`
    :code:`Byte`                :code:`Uint8`               :code:`True`  :code:`Byte` is as general as :code:`Uint8`
    =========================== =========================== ============= ====================================================================

    Args:
        a: The abi.TypeSpec of the value on the right hand side of the assignment.
        b: The abi.TypeSpec of the value on the left hand side of the assignment.

    Returns:
        A boolean result on if type :code:`a` is assignable to type :code:`b`.
    """

    from pyteal.ast.abi import (
        TupleTypeSpec,
        NamedTupleTypeSpec,
        ArrayTypeSpec,
        StaticArrayTypeSpec,
        DynamicArrayTypeSpec,
        StringTypeSpec,
        AddressTypeSpec,
        UintTypeSpec,
    )

    match a, b:
        case NamedTupleTypeSpec(), NamedTupleTypeSpec():
            return a == b
        case TupleTypeSpec(), TupleTypeSpec():
            a, b = cast(TupleTypeSpec, a), cast(TupleTypeSpec, b)
            if a.length_static() != b.length_static():
                return False
            return all(
                map(
                    lambda ab: type_spec_is_assignable_to(ab[0], ab[1]),
                    zip(a.value_type_specs(), b.value_type_specs()),
                )
            )
        case ArrayTypeSpec(), ArrayTypeSpec():
            a, b = cast(ArrayTypeSpec, a), cast(ArrayTypeSpec, b)
            if not type_spec_is_assignable_to(a.value_type_spec(), b.value_type_spec()):
                return False
            match a, b:
                case AddressTypeSpec(), StaticArrayTypeSpec():
                    a, b = cast(AddressTypeSpec, a), cast(StaticArrayTypeSpec, b)
                    return a.length_static() == b.length_static()
                case StaticArrayTypeSpec(), AddressTypeSpec():
                    return False
                case StaticArrayTypeSpec(), StaticArrayTypeSpec():
                    a, b = cast(StaticArrayTypeSpec, a), cast(StaticArrayTypeSpec, b)
                    return a.length_static() == b.length_static()
                case StringTypeSpec(), DynamicArrayTypeSpec():
                    return True
                case DynamicArrayTypeSpec(), StringTypeSpec():
                    return False
                case DynamicArrayTypeSpec(), DynamicArrayTypeSpec():
                    return True
            return False
        case UintTypeSpec(), UintTypeSpec():
            a, b = cast(UintTypeSpec, a), cast(UintTypeSpec, b)
            return a.size == b.size

    if isinstance(a, type(b)):
        return True
    elif str(a) == str(b):
        return True

    return False



from typing import Generic, TypeVar

import pyteal as pt
from pyteal import abi

from tests.blackbox import Blackbox, BlackboxWrapper, PyTealDryRunExecutor

T = TypeVar("T", bound=abi.BaseType)

DEFAULT_DYNAMIC_ARRAY_LENGTH = 3


class ABIRoundtrip(Generic[T]):
    def __init__(
        self,
        annotation_instance: abi.BaseType,
        length: int | None = None,
    ):
        self.instance: abi.BaseType = annotation_instance
        self.type_spec: abi.TypeSpec = annotation_instance.type_spec()
        self.annotation: type[abi.BaseType] = self.type_spec.annotation_type()

        self.length: int | None = length

    def pytealer(self) -> PyTealDryRunExecutor:
        roundtrip = self.roundtrip_factory()
        return PyTealDryRunExecutor(roundtrip, pt.Mode.Application)

    def roundtrip_factory(self) -> BlackboxWrapper:
        comp = self.mutator_factory()

        ann_out = abi.Tuple3[self.annotation, self.annotation, self.annotation]  # type: ignore[misc,name-defined]

        @Blackbox(input_types=[None])
        @pt.ABIReturnSubroutine
        def round_tripper(x: self.annotation, *, output: ann_out):  # type: ignore[name-defined]
            y = abi.make(self.annotation)
            z = abi.make(self.annotation)
            return pt.Seq(y.set(comp(x)), z.set(comp(y)), output.set(x, y, z))  # type: ignore[attr-defined]

        return round_tripper

    def mutator_factory(self) -> pt.ABIReturnSubroutine:
        if isinstance(self.type_spec, abi.BoolTypeSpec):
            return self.bool_comp_factory()
        if isinstance(self.type_spec, abi.UintTypeSpec):
            return self.numerical_comp_factory()
        if isinstance(self.type_spec, abi.StringTypeSpec):
            return self.string_reverse_factory()
        if isinstance(self.type_spec, abi.TupleTypeSpec):
            return self.tuple_comp_factory()
        if isinstance(self.type_spec, abi.ArrayTypeSpec):
            return self.array_comp_factory()
        if isinstance(self.type_spec, abi.TransactionTypeSpec):
            return self.transaction_comp_factory()

        raise ValueError(f"uh-oh!!! didn't handle type {self.instance}")

    def bool_comp_factory(self) -> pt.ABIReturnSubroutine:
        @pt.ABIReturnSubroutine
        def bool_comp(x: abi.Bool, *, output: abi.Bool):
            return output.set(pt.Not(x.get()))

        return bool_comp

    @classmethod
    def max_int(cls, bit_size):
        return (1 << bit_size) - 1

    def transaction_comp_factory(self) -> pt.ABIReturnSubroutine:
        @pt.ABIReturnSubroutine
        def transaction_comp(x: self.annotation, *, output: abi.Uint64):  # type: ignore[name-defined]
            return output.set(x.get().amount())

        return transaction_comp

    def numerical_comp_factory(self) -> pt.ABIReturnSubroutine:
        @pt.ABIReturnSubroutine
        def numerical_comp(x: self.annotation, *, output: self.annotation):  # type: ignore[name-defined]
            max_uint = pt.Int(self.max_int(self.type_spec.bit_size()))  # type: ignore[attr-defined]
            return output.set(max_uint - x.get())

        return numerical_comp

    def string_reverse_factory(self) -> pt.ABIReturnSubroutine:
        """
        Assume strings are python utf-8 compliant and therefore each byte value is at most 127
        """
        if self.length is None:
            self.length = DEFAULT_DYNAMIC_ARRAY_LENGTH

        char_type_spec = abi.ByteTypeSpec()

        @pt.ABIReturnSubroutine
        def string_reverse(x: self.annotation, *, output: self.annotation):  # type: ignore[name-defined]
            insts = [char_type_spec.new_instance() for _ in range(self.length)]  # type: ignore[arg-type]
            setters = [inst.set(x[i]) for i, inst in enumerate(reversed(insts))]
            return pt.Seq(*(setters + [output.set(insts)]))

        return string_reverse

    def tuple_comp_factory(self) -> pt.ABIReturnSubroutine:  # type: ignore[name-defined]
        value_type_specs: list[abi.TypeSpec] = self.type_spec.value_type_specs()  # type: ignore[attr-defined]
        insts = [vts.new_instance() for vts in value_type_specs]
        roundtrips: list[ABIRoundtrip[T]] = [
            ABIRoundtrip(inst, length=None) for inst in insts  # type: ignore[arg-type]
        ]

        @pt.ABIReturnSubroutine
        def tuple_complement(x: self.annotation, *, output: self.annotation):  # type: ignore[name-defined]
            setters = [inst.set(x[i]) for i, inst in enumerate(insts)]  # type: ignore[attr-defined]
            comp_funcs = [rtrip.mutator_factory() for rtrip in roundtrips]
            compers = [inst.set(comp_funcs[i](inst)) for i, inst in enumerate(insts)]  # type: ignore[attr-defined]
            return pt.Seq(*(setters + compers + [output.set(*insts)]))

        return tuple_complement

    def array_comp_factory(self) -> pt.ABIReturnSubroutine:
        """
        When the length has not been provided for a dynamic array,
        default to DEFAULT_DYNAMIC_ARRAY_LENGTH
        """
        if self.length is not None:
            assert self.type_spec.is_length_dynamic()  # type: ignore[attr-defined]
        elif not self.type_spec.is_length_dynamic():  # type: ignore[attr-defined]
            self.length = self.type_spec.length_static()  # type: ignore[attr-defined]
        else:
            self.length = DEFAULT_DYNAMIC_ARRAY_LENGTH

        internal_type_spec = self.type_spec.value_type_spec()  # type: ignore[attr-defined]
        internal_ann_inst = internal_type_spec.new_instance()
        comp_func = ABIRoundtrip(internal_ann_inst, length=None).mutator_factory()

        @pt.ABIReturnSubroutine
        def array_complement(x: self.annotation, *, output: self.annotation):  # type: ignore[name-defined]
            insts = [internal_type_spec.new_instance() for _ in range(self.length)]  # type: ignore[arg-type]
            setters = [inst.set(x[i]) for i, inst in enumerate(insts)]
            compers = [inst.set(comp_func(inst)) for inst in insts]
            return pt.Seq(*(setters + compers + [output.set(insts)]))

        return array_complement



from enum import Enum
import algosdk

from typing import TYPE_CHECKING, cast
from pyteal.ast.int import EnumInt
from pyteal.ast.for_ import For
from pyteal.ast.int import Int
from pyteal.ast.scratchvar import ScratchVar

from pyteal.ast.methodsig import MethodSignature
from pyteal.types import TealType, require_type
from pyteal.errors import TealInputError, TealTypeError, verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr
from pyteal.ast.txn import (
    TxnType,
    TxnArray,
    TxnField,
    TxnExprBuilder,
    TxnaExprBuilder,
    TxnObject,
)
from pyteal.ast.seq import Seq
from pyteal.ast.bytes import Bytes
from pyteal.ast import abi

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class InnerTxnAction(Enum):
    Begin = Op.itxn_begin
    Submit = Op.itxn_submit
    Next = Op.itxn_next


InnerTxnAction.__module__ = "pyteal"


class InnerTxnActionExpr(Expr):
    def __init__(self, action: InnerTxnAction) -> None:
        super().__init__()
        self.action = action

    def __str__(self):
        return "(InnerTxn{})".format(self.action.name)

    def __teal__(self, options: "CompileOptions"):
        op = self.action.value

        verifyProgramVersion(
            op.min_version,
            options.version,
            "Program version too low to create inner transactions",
        )

        return TealBlock.FromOp(options, TealOp(self, op))

    def type_of(self):
        return TealType.none

    def has_return(self):
        return False


class InnerTxnFieldExpr(Expr):
    def __init__(self, field: TxnField, value: Expr) -> None:
        super().__init__()
        require_type(value, field.type_of())
        self.field = field
        self.value = value

    def __str__(self):
        return "(InnerTxnSetField {} {})".format(self.field.arg_name, self.value)

    def __teal__(self, options: "CompileOptions"):
        verifyProgramVersion(
            Op.itxn_field.min_version,
            options.version,
            "Program version too low to create inner transactions",
        )

        return TealBlock.FromOp(
            options, TealOp(self, Op.itxn_field, self.field.arg_name), self.value
        )

    def type_of(self):
        return TealType.none

    def has_return(self):
        return False


class InnerTxnBuilder:
    """This class represents expressions used to create, modify, and submit inner transactions.

    Inner transactions are transactions which applications can dynamically create. Each inner
    transaction will appear as a transaction inside of the current transaction being executed.

    As of program version 5, only the transaction types :any:`TxnType.Payment`, :any:`TxnType.AssetTransfer`,
    :any:`TxnType.AssetConfig`, and :any:`TxnType.AssetFreeze` are allowed. Additionally, not all
    fields are allowed to be set. For example, it is not currently allowed to set the rekeyTo field
    of an inner transaction.
    """

    @classmethod
    def Begin(cls) -> Expr:
        """Begin preparation of a new inner transaction.

        This new inner transaction is initialized with its sender to the application address (:any:`Global.current_application_address`);
        fee to the minimum allowable, taking into account :code:`MinTxnFee` and credit from
        overpaying in earlier transactions; :code:`FirstValid`/:code:`LastValid` to the values in
        the top-level transaction, and all other fields to zero values.

        Requires program version 5 or higher. This operation is only permitted in application mode.
        """
        return InnerTxnActionExpr(InnerTxnAction.Begin)

    @classmethod
    def Next(cls) -> Expr:
        """Begin preparation of a new inner transaction (in the same transaction group).

        This new inner transaction is initialized with its sender to the application address (:any:`Global.current_application_address`);
        fee to the minimum allowable, taking into account :code:`MinTxnFee` and credit from
        overpaying in earlier transactions; :code:`FirstValid`/:code:`LastValid` to the values in
        the top-level transaction, and all other fields to zero values.

        Requires program version 6 or higher. This operation is only permitted in application mode.
        """
        return InnerTxnActionExpr(InnerTxnAction.Next)

    @classmethod
    def Submit(cls) -> Expr:
        """Execute the current inner transaction.

        :any:`InnerTxnBuilder.Begin` and :any:`InnerTxnBuilder.SetField` must be called before
        submitting an inner transaction.

        This will fail if 256 inner transactions have already been executed, or if the
        inner transaction itself fails. Upon failure, the current program will immediately exit and
        fail as well.

        If the inner transaction is successful, then its effects can be immediately observed by this
        program with stateful expressions such as :any:`Balance`. Additionally, the fields of the
        most recently submitted inner transaction can be examined using the :any:`InnerTxn` object.
        If the inner transaction creates an asset, the new asset ID can be found by looking at
        :any:`InnerTxn.created_asset_id() <TxnObject.created_asset_id>`.

        Requires program version 5 or higher. This operation is only permitted in application mode.
        """
        return InnerTxnActionExpr(InnerTxnAction.Submit)

    @classmethod
    def SetField(cls, field: TxnField, value: Expr | list[Expr]) -> Expr:
        """Set a field of the current inner transaction.

        :any:`InnerTxnBuilder.Begin` must be called before setting any fields on an inner
        transaction.

        Note: For non-array field (e.g., note), setting it twice will overwrite the original value.
              While for array field (e.g., accounts), setting it multiple times will append the values.
              It is also possible to pass the entire array field if desired (e.g., Txn.accounts) to pass all the references.

        Requires program version 5 or higher. This operation is only permitted in application mode.

        Args:
            field: The field to set on the inner transaction.
            value: The value to that the field should take. This must evaluate to a type that is
                compatible with the field being set.
        """
        if not field.is_array:
            if type(value) is list or isinstance(value, TxnArray):
                raise TealInputError(
                    "inner transaction set field {} does not support array value".format(
                        field
                    )
                )
            return InnerTxnFieldExpr(field, cast(Expr, value))
        else:
            if type(value) is not list and not isinstance(value, TxnArray):
                raise TealInputError(
                    "inner transaction set array field {} with non-array value".format(
                        field
                    )
                )

            if type(value) is list:
                for valueIter in value:
                    if not isinstance(valueIter, Expr):
                        raise TealInputError(
                            "inner transaction set array field {} with non PyTeal expression array element {}".format(
                                field, valueIter
                            )
                        )

                return Seq(
                    *[
                        InnerTxnFieldExpr(field, cast(Expr, valueIter))
                        for valueIter in value
                    ]
                )
            else:
                arr = cast(TxnArray, value)
                return For(
                    (i := ScratchVar()).store(Int(0)),
                    i.load() < arr.length(),
                    i.store(i.load() + Int(1)),
                ).Do(InnerTxnFieldExpr(field, arr[i.load()]))

    @classmethod
    def Execute(cls, fields: dict[TxnField, Expr | list[Expr]]) -> Expr:
        """Performs a single transaction given fields passed in.

        A convenience method that accepts fields to submit a single inner transaction, which is equivalent to:

        .. code-block:: python

            InnerTxnBuilder.Begin()
            InnerTxnBuilder.SetFields(fields)
            InnerTxnBuilder.Submit()

        Requires program version 5 or higher. This operation is only permitted in application mode.

        Args:
            fields: A dictionary whose keys are fields to set and whose values are the value each
                field should take. Each value must evaluate to a type that is compatible with the
                field being set.
        """
        return Seq(cls.Begin(), cls.SetFields(fields), cls.Submit())

    @classmethod
    def SetFields(cls, fields: dict[TxnField, Expr | list[Expr]]) -> Expr:
        """Set multiple fields of the current inner transaction.

        :any:`InnerTxnBuilder.Begin` must be called before setting any fields on an inner
        transaction.

        Note: For non-array field (e.g., note), setting it twice will overwrite the original value.
              While for array field (e.g., accounts), setting it multiple times will append the values.
              It is also possible to pass the entire array field if desired (e.g., Txn.accounts) to pass all the references.

        Requires program version 5 or higher. This operation is only permitted in application mode.

        Args:
            fields: A dictionary whose keys are fields to set and whose values are the value each
                field should take. Each value must evaluate to a type that is compatible with the
                field being set.
        """
        fieldsToSet = [cls.SetField(field, value) for field, value in fields.items()]
        return Seq(fieldsToSet)

    @classmethod
    def ExecuteMethodCall(
        cls,
        *,
        app_id: Expr,
        method_signature: str,
        args: list[abi.BaseType | Expr | dict[TxnField, Expr | list[Expr]]],
        extra_fields: dict[TxnField, Expr | list[Expr]] = None,
    ) -> Expr:
        """Performs a single app call transaction formatted as an ABI method call.

        A convenience method that accepts fields to submit a single inner transaction, which is equivalent to:

        .. code-block:: python

            InnerTxnBuilder.Begin()
            InnerTxnBuilder.MethodCall(
                app_id=app_id,
                method_signature=method_signature,
                args=args,
                extra_fields=extra_fields,
            ),
            InnerTxnBuilder.Submit()

        Requires program version 5 or higher. This operation is only permitted in application mode.

        Args:
            app_id: An expression that evaluates to a `TealType.uint64` corresponding to the application being called.
            method_signature: A string representing the method signature of the method we're calling. This is used to do
                type checking on the arguments passed and to create the method selector passed as the first argument.
            args: A list of arguments to pass to the application. The values in this list depend on the kind of argument you wish to pass:

                - For basic ABI arguments (not Reference or Transaction types):
                    If an ABI type is passed it **MUST** match the type specified in the `method_signature`. If an Expr is passed it must evaluate to `TealType.bytes` but beyond that no type checking is performed.

                - For Reference arguments:
                    Either the Reference type or an Expr that returns the type corresponding to the reference type are allowed.
                    (i.e. Asset is TealType.uint64, Application is TealType.uint64, Account is TealType.bytes)

                - For Transaction arguments:
                    A dictionary containing TxnField to Expr that describe Transactions to be pre-pended to the transaction group being constructed.  The `TxnField.type_enum` key MUST be set and MUST match the expected transaction type specified in the `method_signature`.

            extra_fields (optional): A dictionary whose keys are fields to set and whose values are the value each
                field should take. Each value must evaluate to a type that is compatible with the
                field being set. These fields are set on the ApplicationCallTransaction being constructed
        """
        return Seq(
            cls.Begin(),
            cls.MethodCall(
                app_id=app_id,
                method_signature=method_signature,
                args=args,
                extra_fields=extra_fields,
            ),
            cls.Submit(),
        )

    @classmethod
    def MethodCall(
        cls,
        *,
        app_id: Expr,
        method_signature: str,
        args: list[abi.BaseType | Expr | dict[TxnField, Expr | list[Expr]]],
        extra_fields: dict[TxnField, Expr | list[Expr]] = None,
    ) -> Expr:
        """Adds an ABI method call transaction to the current inner transaction group.

        :any:`InnerTxnBuilder.Begin` must be called before a MethodCall can be added.

        Requires Teal version 6 or higher. This operation is only permitted in application mode.

        Args:
            app_id: An expression that evaluates to a `TealType.uint64` corresponding to the application being called.
            method_signature: A string representing the method signature of the method we're calling. This is used to do
                type checking on the arguments passed and to create the method selector passed as the first argument.
            args: A list of arguments to pass to the application. The values in this list depend on the kind of argument you wish to pass:

                - For basic ABI arguments (not Reference or Transaction types):
                    If an ABI type is passed it **MUST** match the type specified in the `method_signature`. If an Expr is passed it must evaluate to `TealType.bytes` but beyond that no type checking is performed.

                - For Reference arguments:
                    Either the Reference type or an Expr that returns the type corresponding to the reference type are allowed.
                    (i.e. Asset is TealType.uint64, Application is TealType.uint64, Account is TealType.bytes)

                - For Transaction arguments:
                    A dictionary containing TxnField to Expr that describe Transactions to be pre-pended to the transaction group being constructed.  The `TxnField.type_enum` key MUST be set and MUST match the expected transaction type specified in the `method_signature`.

            extra_fields (optional): A dictionary whose keys are fields to set and whose values are the value each
                field should take. Each value must evaluate to a type that is compatible with the
                field being set. These fields are set on the ApplicationCallTransaction being constructed
        """

        from pyteal.ast.abi.util import type_spec_is_assignable_to

        require_type(app_id, TealType.uint64)

        # Default, always need these
        fields_to_set = [
            cls.SetField(TxnField.type_enum, TxnType.ApplicationCall),
            cls.SetField(TxnField.application_id, app_id),
        ]

        # We only care about the args
        arg_type_specs: list[abi.TypeSpec]
        arg_type_specs, _ = abi.type_specs_from_signature(method_signature)

        if len(args) != len(arg_type_specs):
            raise TealInputError(
                f"Expected {len(arg_type_specs)} arguments, got {len(args)}"
            )

        # Start app args with the method selector
        app_args: list[Expr] = [MethodSignature(method_signature)]

        # Transactions are not included in the App Call
        txns_to_pass: list[Expr] = []

        # Reference Types are treated specially
        accts: list[Expr] = []
        apps: list[Expr] = []
        assets: list[Expr] = []

        for idx, method_arg_ts in enumerate(arg_type_specs):
            arg = args[idx]

            if method_arg_ts in abi.TransactionTypeSpecs:
                if not isinstance(arg, dict):
                    raise TealTypeError(arg, dict[TxnField, Expr | list[Expr]])

                if TxnField.type_enum not in arg:
                    raise TealInputError(
                        f"Expected Transaction at arg {idx} to contain field type_enum"
                    )

                if type(arg[TxnField.type_enum]) is not EnumInt:
                    raise TealTypeError(arg[TxnField.type_enum], EnumInt)

                txntype = cast(EnumInt, arg[TxnField.type_enum]).name
                # If the arg is an unspecified transaction, no need to check the type_enum

                if not type_spec_is_assignable_to(
                    abi.type_spec_from_algosdk(txntype), method_arg_ts
                ):
                    raise TealInputError(
                        f"Expected Transaction at arg {idx} to be {method_arg_ts}, got {txntype}"
                    )

                txns_to_pass.append(InnerTxnBuilder.SetFields(arg))

            elif method_arg_ts in abi.ReferenceTypeSpecs:
                match method_arg_ts:
                    # For both acct and application, add index to
                    # app args _after_ appending since 0 is implicitly set
                    case abi.AccountTypeSpec():
                        if isinstance(arg, Expr):
                            # require the address is passed
                            require_type(arg, TealType.bytes)
                            accts.append(arg)
                        elif isinstance(arg, abi.Account):
                            accts.append(arg.address())
                        else:
                            raise TealTypeError(arg, abi.Account | Expr)

                        app_args.append(
                            Bytes(
                                algosdk.abi.ABIType.from_string("uint8").encode(
                                    len(accts)
                                )
                            )
                        )

                    case abi.ApplicationTypeSpec():
                        if isinstance(arg, Expr):
                            # require the app id be passed
                            require_type(arg, TealType.uint64)
                            apps.append(arg)
                        elif isinstance(arg, abi.Application):
                            apps.append(arg.application_id())
                        else:
                            raise TealTypeError(arg, abi.Application | Expr)

                        app_args.append(
                            Bytes(
                                algosdk.abi.ABIType.from_string("uint8").encode(
                                    len(apps)
                                )
                            )
                        )

                    # For assets, add to app_args prior to appending to assets array
                    case abi.AssetTypeSpec():
                        app_args.append(
                            Bytes(
                                algosdk.abi.ABIType.from_string("uint8").encode(
                                    len(assets)
                                )
                            )
                        )

                        if isinstance(arg, Expr):
                            require_type(arg, TealType.uint64)
                            assets.append(arg)
                        elif isinstance(arg, abi.Asset):
                            assets.append(arg.asset_id())
                        else:
                            raise TealTypeError(arg, abi.Asset | Expr)
            else:
                if isinstance(arg, Expr):
                    # This should _always_ be bytes, since we assume its already abi encoded
                    require_type(arg, TealType.bytes)
                    app_args.append(arg)
                elif isinstance(arg, abi.BaseType):
                    if not type_spec_is_assignable_to(arg.type_spec(), method_arg_ts):
                        raise TealTypeError(arg.type_spec(), method_arg_ts)
                    app_args.append(arg.encode())
                else:
                    raise TealTypeError(arg, abi.BaseType | Expr)

        if len(accts) > 0:
            fields_to_set.append(cls.SetField(TxnField.accounts, accts))

        if len(apps) > 0:
            fields_to_set.append(cls.SetField(TxnField.applications, apps))

        if len(assets) > 0:
            fields_to_set.append(cls.SetField(TxnField.assets, assets))

        fields_to_set.append(cls.SetField(TxnField.application_args, app_args))

        return Seq(
            # Add the transactions first
            *[Seq(ttp, InnerTxnBuilder.Next()) for ttp in txns_to_pass],
            # Set the fields for the app call in app args and foreign arrays
            *fields_to_set,
            # Add any remaining fields specified by the user
            InnerTxnBuilder.SetFields({} if extra_fields is None else extra_fields),
        )


InnerTxnBuilder.__module__ = "pyteal"

InnerTxn: TxnObject = TxnObject(
    TxnExprBuilder(Op.itxn, "InnerTxn"),
    TxnaExprBuilder(Op.itxna, Op.itxnas, "InnerTxna"),
)

InnerTxn.__module__ = "pyteal"
